# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log as oslo_logging

import masakarimonitors.conf
from masakarimonitors.ha import masakari


LOG = oslo_logging.getLogger(__name__)
CONF = masakarimonitors.conf.CONF


class Callback(object):
    """Class of callback processing."""

    def __init__(self):
        self.notifier = masakari.SendNotification()

    def libvirt_event_callback(self, event_id, details, domain_uuid,
                               notice_type, hostname, current_time):
        """Callback method.

        Callback processing be executed as result of the
        libvirt event filter.

        :param event_id: Event ID notify to the callback function
        :param details: Event code notify to the callback function
        :param domain_uuid: Uuid notify to the callback function
        :param notice_type: Notice type notify to the callback function
        :param hostname: Server host name of the source an event occur
                         notify to the callback function
        :param current_time: Event occurred time notify to the callback
                            function
        """

        # Output to the syslog.
        LOG.info("Libvirt Event: type=%(notice_type)s,"
                 " hostname=%(hostname)s,"
                 " uuid=%(uuid)s, time=%(current_time)s,"
                 " event_id=%(event_id)s,"
                 " detail=%(detail)s)" % {'notice_type': notice_type,
                                          'hostname': hostname,
                                          'uuid': domain_uuid,
                                          'current_time': current_time,
                                          'event_id': event_id,
                                          'detail': details})

        # Set the event to the dictionary.
        event = {
            'notification': {
                'type': notice_type,
                'hostname': hostname,
                'generated_time': current_time,
                'payload': {
                    'event': event_id,
                    'instance_uuid': domain_uuid,
                    'vir_domain_event': details
                }
            }
        }

        self.notifier.send_notification(CONF.callback.retry_max,
                                        CONF.callback.retry_interval,
                                        event)



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pbr import version as pbr_version

MONITORS_VENDOR = "OpenStack Foundation"
MONITORS_PRODUCT = "OpenStack Masakari Monitors"
MONITORS_PACKAGE = None  # OS distro package version suffix

loaded = False
version_info = pbr_version.VersionInfo('masakari-monitors')
version_string = version_info.version_string


def _load_config():
    # Don't load in global context, since we can't assume
    # these modules are accessible when distutils uses
    # this module
    import configparser

    from oslo_config import cfg

    from oslo_log import log as logging

    global loaded, MONITORS_VENDOR, MONITORS_PRODUCT, MONITORS_PACKAGE
    if loaded:
        return

    loaded = True

    cfgfile = cfg.CONF.find_file("release")
    if cfgfile is None:
        return

    try:
        cfg = configparser.RawConfigParser()
        cfg.read(cfgfile)

        if cfg.has_option("Masakarimonitors", "vendor"):
            MONITORS_VENDOR = cfg.get("Masakarimonitors", "vendor")

        if cfg.has_option("Masakarimonitors", "product"):
            MONITORS_PRODUCT = cfg.get("Masakarimonitors", "product")

        if cfg.has_option("Masakarimonitors", "package"):
            MONITORS_PACKAGE = cfg.get("Masakarimonitors", "package")
    except Exception as ex:
        LOG = logging.getLogger(__name__)
        LOG.error("Failed to load %(cfgfile)s: %(ex)s",
                  {'cfgfile': cfgfile, 'ex': ex})


def vendor_string():
    _load_config()

    return MONITORS_VENDOR


def product_string():
    _load_config()

    return MONITORS_PRODUCT


def package_string():
    _load_config()

    return MONITORS_PACKAGE


def version_string_with_package():
    if package_string() is None:
        return version_info.version_string()
    else:
        return "%s-%s" % (version_info.version_string(), package_string())



# Copyright(c) 2021 Inspur
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import testtools
from unittest import mock

from collections import deque
import eventlet
from oslo_config import fixture as fixture_config

import masakarimonitors.conf
from masakarimonitors.ha import masakari
from masakarimonitors.hostmonitor.consul_check import consul_helper
from masakarimonitors.hostmonitor.consul_check import manager
from masakarimonitors.hostmonitor.consul_check import matrix_helper

eventlet.monkey_patch(os=False)

CONF = masakarimonitors.conf.CONF


class TestConsulCheck(testtools.TestCase):

    def setUp(self):
        super(TestConsulCheck, self).setUp()
        self.CONF = self.useFixture(fixture_config.Config()).conf
        self.host_monitor = manager.ConsulCheck()
        self.host_monitor.matrix_manager = \
            matrix_helper.MatrixManager(self.CONF)
        self.host_monitor._matrix = matrix_helper.DEFAULT_MATRIX
        self.host_monitor.consul_manager = \
            consul_helper.ConsulManager(self.CONF)
        self.host_monitor._sequence = ['manage', 'tenant', 'storage']
        self.host_monitor.monitoring_data = {
            "node01": deque([['up', 'up', 'up'],
                             ['up', 'up', 'up'],
                             ['up', 'up', 'up']],
                            maxlen=3),
            "node02": deque([['up', 'up', 'up'],
                             ['up', 'up', 'up'],
                             ['up', 'up', 'down']],
                            maxlen=3),
            "node03": deque([['up', 'up', 'up'],
                             ['down', 'up', 'up'],
                             ['down', 'up', 'up']],
                            maxlen=3),
        }

    def test_update_monitoring_data(self):
        mock_health = {
            'node01': ['up', 'up', 'up'],
            'node02': ['up', 'up', 'up'],
            'node03': ['up', 'up', 'up']
        }

        with mock.patch.object(self.host_monitor.consul_manager, 'get_health',
                return_value=mock_health):
            self.host_monitor.update_monitoring_data()
            excepted_monitoring_data = {
                "node01": deque([['up', 'up', 'up'],
                                 ['up', 'up', 'up'],
                                 ['up', 'up', 'up']],
                                maxlen=3),
                "node02": deque([['up', 'up', 'up'],
                                 ['up', 'up', 'down'],
                                 ['up', 'up', 'up']],
                                maxlen=3),
                "node03": deque([['down', 'up', 'up'],
                                 ['down', 'up', 'up'],
                                 ['up', 'up', 'up']],
                                maxlen=3),
            }
            self.assertEqual(excepted_monitoring_data,
                             self.host_monitor.monitoring_data)

    def test_get_host_statistical_health(self):
        self.assertEqual(['up', 'up', 'up'],
                         self.host_monitor.get_host_health('node01'))
        self.assertEqual(['up', 'up', None],
                         self.host_monitor.get_host_health('node02'))
        self.assertEqual([None, 'up', 'up'],
                         self.host_monitor.get_host_health('node03'))

    def test_host_statistical_health_changed(self):
        self.host_monitor.last_host_health = {
            'node02': ['up', 'up', None],
            'node03': ['up', 'up', 'down']
        }

        self.assertFalse(self.host_monitor._host_health_changed(
            'node01', ['up', 'up', 'up']))
        self.assertTrue(self.host_monitor._host_health_changed(
            'node02', ['up', 'up', 'up']))
        self.assertTrue(self.host_monitor._host_health_changed(
            'node03', ['up', 'up', 'up']))
        last_host_health = {
            'node01': ['up', 'up', 'up'],
            'node02': ['up', 'up', 'up'],
            'node03': ['up', 'up', 'up']
        }
        self.assertEqual(self.host_monitor.last_host_health,
                         last_host_health)

    def test_get_action_from_matrix_by_host_health(self):
        self.assertEqual(
            [],
            self.host_monitor.get_action_from_matrix(['up', 'up', 'up']))
        self.assertEqual(
            ["recovery"],
            self.host_monitor.get_action_from_matrix(['up', 'up', 'down']))
        self.assertEqual(
            [],
            self.host_monitor.get_action_from_matrix(['down', 'up', 'up']))
        self.assertEqual(
            ["recovery"],
            self.host_monitor.get_action_from_matrix(['down', 'down', 'down']))

    @mock.patch.object(masakari.SendNotification, 'send_notification')
    @mock.patch.object(manager.ConsulCheck, '_event')
    def test_poll_hosts(self, mock_event, mock_send_notification):
        self.host_monitor.monitoring_data = {
            "node01": deque([['up', 'up', 'up'],
                             ['up', 'up', 'up'],
                             ['up', 'up', 'up']],
                            maxlen=3),
            "node02": deque([['up', 'up', 'down'],
                             ['up', 'up', 'down'],
                             ['up', 'up', 'down']],
                            maxlen=3),
            "node03": deque([['up', 'up', 'up'],
                             ['up', 'up', 'up'],
                             ['up', 'up', 'up']],
                            maxlen=3),
        }

        self.host_monitor.last_host_health = {
            'node02': ['up', 'up', None],
            'node03': ['up', 'up', 'up']
        }

        test_event = {'notification': 'test'}
        mock_event.return_value = test_event

        self.host_monitor.poll_hosts()
        mock_send_notification.assert_called_once_with(
            CONF.host.api_retry_max, CONF.host.api_retry_interval, test_event)



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import libvirt

# If is not defined internal , -1 is stored.
DUMMY = -1

# Enumerate all event that can get.
# Comment out events that is not targeted in the callback.
event_filter_dic = {
    libvirt.VIR_DOMAIN_EVENT_ID_LIFECYCLE:
    {
        libvirt.VIR_DOMAIN_EVENT_SUSPENDED:
        (
            libvirt.VIR_DOMAIN_EVENT_SUSPENDED_IOERROR,
            libvirt.VIR_DOMAIN_EVENT_SUSPENDED_WATCHDOG,
            libvirt.VIR_DOMAIN_EVENT_SUSPENDED_API_ERROR
        ),
        libvirt.VIR_DOMAIN_EVENT_STOPPED:
        (
            libvirt.VIR_DOMAIN_EVENT_STOPPED_SHUTDOWN,
            libvirt.VIR_DOMAIN_EVENT_STOPPED_DESTROYED,
            libvirt.VIR_DOMAIN_EVENT_STOPPED_FAILED,
        ),
        libvirt.VIR_DOMAIN_EVENT_SHUTDOWN:
        (
            libvirt.VIR_DOMAIN_EVENT_SHUTDOWN_FINISHED,
        )
    },
    libvirt.VIR_DOMAIN_EVENT_ID_REBOOT: {DUMMY: (DUMMY,)},
    libvirt.VIR_DOMAIN_EVENT_ID_WATCHDOG:
    {
        libvirt.VIR_DOMAIN_EVENT_WATCHDOG_NONE: (DUMMY,),
        libvirt.VIR_DOMAIN_EVENT_WATCHDOG_PAUSE: (DUMMY,),
        libvirt.VIR_DOMAIN_EVENT_WATCHDOG_RESET: (DUMMY,),
        libvirt.VIR_DOMAIN_EVENT_WATCHDOG_POWEROFF: (DUMMY,),
        libvirt.VIR_DOMAIN_EVENT_WATCHDOG_SHUTDOWN: (DUMMY,),
        libvirt.VIR_DOMAIN_EVENT_WATCHDOG_DEBUG: (DUMMY,)
    },
    libvirt.VIR_DOMAIN_EVENT_ID_IO_ERROR:
    {
        libvirt.VIR_DOMAIN_EVENT_IO_ERROR_NONE: (DUMMY,),
        libvirt.VIR_DOMAIN_EVENT_IO_ERROR_PAUSE: (DUMMY,),
        libvirt.VIR_DOMAIN_EVENT_IO_ERROR_REPORT: (DUMMY,)
    },
    libvirt.VIR_DOMAIN_EVENT_ID_IO_ERROR_REASON: {DUMMY: (DUMMY,)},
    libvirt.VIR_DOMAIN_EVENT_ID_CONTROL_ERROR: {DUMMY: (DUMMY,)}

}

eventID_dic = {
    libvirt.VIR_DOMAIN_EVENT_ID_LIFECYCLE: 'LIFECYCLE',
    libvirt.VIR_DOMAIN_EVENT_ID_REBOOT: 'REBOOT',
    libvirt.VIR_DOMAIN_EVENT_ID_WATCHDOG: 'WATCHDOG',
    libvirt.VIR_DOMAIN_EVENT_ID_IO_ERROR: 'IO_ERROR',
    libvirt.VIR_DOMAIN_EVENT_ID_IO_ERROR_REASON: 'IO_ERROR_REASON',
    libvirt.VIR_DOMAIN_EVENT_ID_CONTROL_ERROR: 'CONTROL_ERROR'}

detail_dic = {
    libvirt.VIR_DOMAIN_EVENT_ID_LIFECYCLE: {
        libvirt.VIR_DOMAIN_EVENT_SUSPENDED: {
            libvirt.VIR_DOMAIN_EVENT_SUSPENDED_IOERROR:
            'SUSPENDED_IOERROR',
            libvirt.VIR_DOMAIN_EVENT_SUSPENDED_WATCHDOG:
            'SUSPENDED_WATCHDOG',
            libvirt.VIR_DOMAIN_EVENT_SUSPENDED_API_ERROR:
            'SUSPENDED_API_ERROR'},
        libvirt.VIR_DOMAIN_EVENT_STOPPED: {
            libvirt.VIR_DOMAIN_EVENT_STOPPED_SHUTDOWN:
            'STOPPED_SHUTDOWN',
            libvirt.VIR_DOMAIN_EVENT_STOPPED_DESTROYED:
            'STOPPED_DESTROYED',
            libvirt.VIR_DOMAIN_EVENT_STOPPED_FAILED:
            'STOPPED_FAILED'},
        libvirt.VIR_DOMAIN_EVENT_SHUTDOWN: {
            libvirt.VIR_DOMAIN_EVENT_SHUTDOWN_FINISHED:
            'SHUTDOWN_FINISHED'}
    },
    libvirt.VIR_DOMAIN_EVENT_ID_REBOOT: {
        DUMMY: {
            DUMMY: 'UNKNOWN'}},
    libvirt.VIR_DOMAIN_EVENT_ID_WATCHDOG: {
        libvirt.VIR_DOMAIN_EVENT_WATCHDOG_NONE: {
            DUMMY: 'WATCHDOG_NONE'},
        libvirt.VIR_DOMAIN_EVENT_WATCHDOG_PAUSE: {
            DUMMY: 'WATCHDOG_PAUSE'},
        libvirt.VIR_DOMAIN_EVENT_WATCHDOG_RESET: {
            DUMMY: 'WATCHDOG_RESET'},
        libvirt.VIR_DOMAIN_EVENT_WATCHDOG_POWEROFF: {
            DUMMY: 'WATCHDOG_POWEROFF'},
        libvirt.VIR_DOMAIN_EVENT_WATCHDOG_SHUTDOWN: {
            DUMMY: 'WATCHDOG_SHUTDOWN'},
        libvirt.VIR_DOMAIN_EVENT_WATCHDOG_DEBUG: {
            DUMMY: 'WATCHDOG_DEBUG'}},
    libvirt.VIR_DOMAIN_EVENT_ID_IO_ERROR: {
        libvirt.VIR_DOMAIN_EVENT_IO_ERROR_NONE: {
            DUMMY: 'IO_ERROR_NONE'},
        libvirt.VIR_DOMAIN_EVENT_IO_ERROR_PAUSE: {
            DUMMY: 'IO_ERROR_PAUSE'},
        libvirt.VIR_DOMAIN_EVENT_IO_ERROR_REPORT: {
            DUMMY: 'IO_ERROR_REPORT'}},
    libvirt.VIR_DOMAIN_EVENT_ID_IO_ERROR_REASON: {
        DUMMY: {
            DUMMY: 'UNKNOWN'}},
    libvirt.VIR_DOMAIN_EVENT_ID_CONTROL_ERROR: {
        DUMMY: {
            DUMMY: 'UNKNOWN'}}
}



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import libvirt
import testtools
import threading
import time
from unittest import mock
import uuid

import eventlet

from masakarimonitors.instancemonitor import instance
from masakarimonitors.instancemonitor.libvirt_handler import eventfilter

eventlet.monkey_patch(os=False)


class TestInstancemonitorManager(testtools.TestCase):

    def setUp(self):
        super(TestInstancemonitorManager, self).setUp()

    def _make_callback_params(self):
        mock_conn = mock.Mock()
        mock_dom = mock.Mock()
        test_uuid = uuid.uuid4()
        mock_dom.UUIDString.return_value = test_uuid
        mock_opaque = mock.Mock()

        return mock_conn, mock_dom, mock_opaque, test_uuid

    @mock.patch.object(libvirt, 'virEventRunDefaultImpl')
    def test_vir_event_loop_native_run(self, mock_virEventRunDefaultImpl):
        mock_virEventRunDefaultImpl.side_effect = Exception("Test exception.")

        obj = instance.InstancemonitorManager()
        exception_flag = False
        try:
            obj._vir_event_loop_native_run()
        except Exception:
            exception_flag = True

        self.assertTrue(exception_flag)
        mock_virEventRunDefaultImpl.assert_called_once()

    @mock.patch.object(eventfilter.EventFilter, 'vir_event_filter')
    def test_my_domain_event_callback(self, mock_vir_event_filter):
        mock_vir_event_filter.return_value = None
        mock_conn, mock_dom, mock_opaque, test_uuid = \
            self._make_callback_params()

        obj = instance.InstancemonitorManager()
        obj._my_domain_event_callback(mock_conn, mock_dom, 0, 1, mock_opaque)

        mock_vir_event_filter.assert_called_once_with(
            libvirt.VIR_DOMAIN_EVENT_ID_LIFECYCLE, 0, 1, test_uuid)

    @mock.patch.object(eventfilter.EventFilter, 'vir_event_filter')
    def test_my_domain_event_reboot_callback(self, mock_vir_event_filter):
        mock_vir_event_filter.return_value = None
        mock_conn, mock_dom, mock_opaque, test_uuid = \
            self._make_callback_params()

        obj = instance.InstancemonitorManager()
        obj._my_domain_event_reboot_callback(mock_conn, mock_dom, mock_opaque)

        mock_vir_event_filter.assert_called_once_with(
            libvirt.VIR_DOMAIN_EVENT_ID_REBOOT, -1, -1, test_uuid)

    @mock.patch.object(eventfilter.EventFilter, 'vir_event_filter')
    def test_my_domain_event_rtc_change_callback(self, mock_vir_event_filter):
        mock_vir_event_filter.return_value = None
        mock_conn, mock_dom, mock_opaque, test_uuid = \
            self._make_callback_params()
        utcoffset = ""

        obj = instance.InstancemonitorManager()
        obj._my_domain_event_rtc_change_callback(
            mock_conn, mock_dom, utcoffset, mock_opaque)

        mock_vir_event_filter.assert_called_once_with(
            libvirt.VIR_DOMAIN_EVENT_ID_RTC_CHANGE, -1, -1, test_uuid)

    @mock.patch.object(eventfilter.EventFilter, 'vir_event_filter')
    def test_my_domain_event_watchdog_callback(self, mock_vir_event_filter):
        mock_vir_event_filter.return_value = None
        mock_conn, mock_dom, mock_opaque, test_uuid = \
            self._make_callback_params()
        action = 0

        obj = instance.InstancemonitorManager()
        obj._my_domain_event_watchdog_callback(
            mock_conn, mock_dom, action, mock_opaque)

        mock_vir_event_filter.assert_called_once_with(
            libvirt.VIR_DOMAIN_EVENT_ID_WATCHDOG, action, -1, test_uuid)

    @mock.patch.object(eventfilter.EventFilter, 'vir_event_filter')
    def test_my_domain_event_io_error_callback(self, mock_vir_event_filter):
        mock_vir_event_filter.return_value = None
        mock_conn, mock_dom, mock_opaque, test_uuid = \
            self._make_callback_params()
        srcpath = ""
        devalias = ""
        action = 0

        obj = instance.InstancemonitorManager()
        obj._my_domain_event_io_error_callback(
            mock_conn, mock_dom, srcpath, devalias, action, mock_opaque)

        mock_vir_event_filter.assert_called_once_with(
            libvirt.VIR_DOMAIN_EVENT_ID_IO_ERROR, action, -1, test_uuid)

    @mock.patch.object(eventfilter.EventFilter, 'vir_event_filter')
    def test_my_domain_event_graphics_callback(self, mock_vir_event_filter):
        mock_vir_event_filter.return_value = None
        mock_conn, mock_dom, mock_opaque, test_uuid = \
            self._make_callback_params()
        phase = 0
        localAddr = ""
        remoteAddr = ""
        authScheme = ""
        subject = ""

        obj = instance.InstancemonitorManager()
        obj._my_domain_event_graphics_callback(
            mock_conn, mock_dom, phase, localAddr, remoteAddr, authScheme,
            subject, mock_opaque)

        mock_vir_event_filter.assert_called_once_with(
            libvirt.VIR_DOMAIN_EVENT_ID_GRAPHICS, -1, phase, test_uuid)

    @mock.patch.object(eventfilter.EventFilter, 'vir_event_filter')
    def test_my_domain_event_disk_change_callback(
            self, mock_vir_event_filter):
        mock_vir_event_filter.return_value = None
        mock_conn, mock_dom, mock_opaque, test_uuid = \
            self._make_callback_params()
        oldSrcPath = ""
        newSrcPath = ""
        devAlias = ""
        reason = ""

        obj = instance.InstancemonitorManager()
        obj._my_domain_event_disk_change_callback(
            mock_conn, mock_dom, oldSrcPath, newSrcPath, devAlias, reason,
            mock_opaque)

        mock_vir_event_filter.assert_called_once_with(
            libvirt.VIR_DOMAIN_EVENT_ID_DISK_CHANGE, -1, -1, test_uuid)

    @mock.patch.object(eventfilter.EventFilter, 'vir_event_filter')
    def test_my_domain_event_io_error_reason_callback(
            self, mock_vir_event_filter):
        mock_vir_event_filter.return_value = None
        mock_conn, mock_dom, mock_opaque, test_uuid = \
            self._make_callback_params()
        srcPath = ""
        devAlias = ""
        action = ""
        reason = ""

        obj = instance.InstancemonitorManager()
        obj._my_domain_event_io_error_reason_callback(
            mock_conn, mock_dom, srcPath, devAlias, action, reason,
            mock_opaque)

        mock_vir_event_filter.assert_called_once_with(
            libvirt.VIR_DOMAIN_EVENT_ID_IO_ERROR_REASON, -1, -1, test_uuid)

    @mock.patch.object(eventfilter.EventFilter, 'vir_event_filter')
    def test_my_domain_event_generic_callback(self, mock_vir_event_filter):
        mock_vir_event_filter.return_value = None
        mock_conn, mock_dom, mock_opaque, test_uuid = \
            self._make_callback_params()

        obj = instance.InstancemonitorManager()
        obj._my_domain_event_generic_callback(
            mock_conn, mock_dom, mock_opaque)

        mock_vir_event_filter.assert_called_once_with(
            libvirt.VIR_DOMAIN_EVENT_ID_CONTROL_ERROR, -1, -1, test_uuid)

    def test_err_handler(self):
        obj = instance.InstancemonitorManager()
        obj._err_handler("Test context.", ('err0.', 'err1', 'err2'))

    def test_stop(self):
        obj = instance.InstancemonitorManager()
        obj.stop()
        self.assertFalse(obj.running)

    @mock.patch.object(time, 'sleep')
    @mock.patch.object(eventlet.greenthread, 'sleep')
    @mock.patch.object(libvirt, 'openAuth')
    @mock.patch.object(threading, 'Thread')
    @mock.patch.object(libvirt, 'virEventRegisterDefaultImpl')
    def test_main(self,
                  mock_virEventRegisterDefaultImpl,
                  mock_Thread,
                  mock_openAuth,
                  mock_greenthread_sleep,
                  mock_time_sleep):

        mock_virEventRegisterDefaultImpl.return_value = None
        mock_event_loop_thread = mock.Mock(return_value=None)
        mock_Thread.return_value = mock_event_loop_thread
        mock_vc = mock.Mock()
        mock_openAuth.return_value = mock_vc
        mock_vc.domainEventRegisterAny.side_effect = \
            [0, 0, 0, 0, 0, 0, 0, 0, 0]
        mock_vc.setKeepAlive.return_value = None
        mock_vc.isAlive.side_effect = [1, 0]
        mock_vc.domainEventDeregisterAny.side_effect = \
            [None, None, None, None, None, None, None, None,
             Exception("Test exception.")]
        mock_vc.close.return_value = None
        mock_greenthread_sleep.return_value = None
        mock_time_sleep.side_effect = Exception("Test exception.")

        obj = instance.InstancemonitorManager()
        exception_flag = False
        try:
            obj.main()
        except Exception:
            exception_flag = True

        handlers_count = 9
        self.assertTrue(exception_flag)
        mock_virEventRegisterDefaultImpl.assert_called_once()
        self.assertEqual(True, mock_event_loop_thread.daemon)
        mock_event_loop_thread.start.assert_called_once()
        mock_openAuth.assert_called_once_with(
            "qemu:///system",
            [[2, 6, 8, 5, 7, 9],
             instance.InstancemonitorManager._connect_auth_cb,
             None], 1)
        self.assertEqual(
            handlers_count, mock_vc.domainEventRegisterAny.call_count)
        mock_vc.setKeepAlive.assert_called_once_with(5, 3)
        self.assertEqual(2, mock_vc.isAlive.call_count)
        self.assertEqual(
            handlers_count, mock_vc.domainEventDeregisterAny.call_count)
        mock_vc.close.assert_called_once()



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from keystoneauth1 import loading as ks_loading
from oslo_config import cfg

API_GROUP = 'api'

api_group = cfg.OptGroup(
    API_GROUP,
    title='Api Options',
    help="""
Configuration options for sending notifications.
""")

monitor_api_opts = [
    cfg.StrOpt('region',
               default='RegionOne',
               help='Region name.'),
    cfg.StrOpt('api_version',
               default='v1',
               help='Masakari API Version.'),
    cfg.StrOpt('api_interface',
               default='public',
               help='Interface of endpoint.'),
]


def register_opts(conf):
    conf.register_group(api_group)
    conf.register_opts(monitor_api_opts, group=api_group)
    ks_loading.register_session_conf_options(conf, api_group)
    ks_loading.register_auth_conf_options(conf, api_group)
    conf.set_default('auth_type', 'password', group=api_group)


def list_opts():
    return {
        api_group: (
            monitor_api_opts +
            ks_loading.get_auth_plugin_conf_options('password'))
    }



# Copyright (c) 2018 WindRiver Systems
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Introspective instance monitoring depends on the qemu guest agent to
# monitoring what inside of a VM.
#
# A few items to get in the way of the design:
#
# - After VM is active, it needs time to start qemu-guest-agent.
# Before error/failure is reported, we need a discovery phase to wait
# until VM is guest_pingable.
#
# - Debouncing is needed to enforce that masakari_notifier function
# not calling twice or more for the same failure.
#
# After reported a masakari notification, sent_notification flag will
# need to be reset when there is a QEMU LIFECYCLE event like:
#  STARTED_BOOTED
#  SUSPENDED_PAUSED

import eventlet
import libvirt
import libvirtmod_qemu
import logging
import re
import socket
import time
import traceback

# Machine module contains the state machine logic, state and transition events
from automaton import machines
from dogpile.cache.api import NoValue
from lxml import etree
from oslo_config import cfg
from oslo_utils import timeutils

from masakarimonitors.ha import masakari
from masakarimonitors.introspectiveinstancemonitor import cache
from masakarimonitors.objects import event_constants as ec
from masakarimonitors import utils

CONF = cfg.CONF
ICONF = cfg.CONF.introspectiveinstancemonitor

# The VM QEMU Quest Agent states
#
# discovery = initial state of VM,
#   remains in this state until it is determined that
#   the VM has a qemu-agent interface enabling intrusive-instance-monitoring
#
# healthy = no failure event is detected
#
# error = An error is recorded on every audit cycle where
#  the VM is in the error state
#
# reported = a transient state
#   to keep track of reporting of notification
#

# Transitions
#
#   Representation of a transition managed by a ``Machine`` instance.
#   source (str): Source state of the transition.
#   dest (str): Destination state of the transition.
#   trigger (str): The type of triggering event
#     that advances to the next state in the sequence.


def action_on_enter(new_state, triggered_event):
    pass


def action_on_exit(old_state, triggered_event):
    pass


STATE_SPACE = [
    {
        'name': 'discovery',
        'next_states': {
            'guest_pingable': 'healthy',
            'guest_not_pingable': 'discovery',
        },
        'on_enter': action_on_enter,
        'on_exit': action_on_exit,
    },
    {
        'name': 'healthy',
        'next_states': {
            'guest_pingable': 'healthy',
            'guest_not_pingable': 'error',
        },
        'on_enter': action_on_enter,
        'on_exit': action_on_exit,
    },
    {
        'name': 'error',
        'next_states': {
            'report': 'reported',
            'guest_pingable': 'healthy',
            'guest_not_pingable': 'error',
        },
        'on_enter': action_on_enter,
        'on_exit': action_on_exit,
    },
    {
        'name': 'reported',
        'next_states': {
            'guest_pingable': 'discovery',
            'guest_not_pingable': 'reported',
        },
        'on_enter': action_on_enter,
        'on_exit': action_on_exit,
    },
]


#   Journal representation of a managed ``Machine`` instance.
class Journal(machines.FiniteMachine):

    # Factory pattern to create a Journal object
    @classmethod
    def factory(cls, domain_uuid):
        jo = cls.build(STATE_SPACE)
        jo.default_start_state = 'discovery'
        jo.initialize()
        jo.failedCount = 0
        # Conditions to reset sent_notification
        jo.sent_notification = False
        jo.domain_uuid = domain_uuid
        jo.lastUsed = time.time()
        LOG.debug(str(domain_uuid) + ':Journal:__init__:')
        return jo

    def resetState(self):
        self.default_start_state = 'discovery'
        self.initialize()
        self.failedCount = 0
        self.sent_notification = False
        LOG.debug(str(self.domain_uuid) + '__resetState__:')

    def processEvent(self, event):
        self.process_event(event)
        self.lastUsed = time.time()

    def getFailedCount(self):
        return self.failedCount

    def incrementFailedCount(self):
        if (self.current_state == 'error'):
            self.failedCount += 1

    def resetFailedCount(self):
        self.failedCount = 0

    def setSentNotification(self, boolean):
        self.sent_notification = boolean

    def getSentNotification(self):
        return self.sent_notification


# libvirt state verbose dictionary
STATES = {
    libvirt.VIR_DOMAIN_NOSTATE: 'no state',
    libvirt.VIR_DOMAIN_RUNNING: 'running',
    libvirt.VIR_DOMAIN_BLOCKED: 'blocked on resource',
    libvirt.VIR_DOMAIN_PAUSED: 'paused by user',
    libvirt.VIR_DOMAIN_SHUTDOWN: 'being shut down',
    libvirt.VIR_DOMAIN_SHUTOFF: 'shut off',
    libvirt.VIR_DOMAIN_CRASHED: 'crashed',
}

LOG = logging.getLogger(__name__)


def get_function_name():
    return traceback.extract_stack(None, 2)[0][2]


# To reset journal object by domain uuid
def resetJournal(domain_uuid):
    # To get the VM Journal object from the dictionary
    #    :param domain: QEMU domain UUID
    dict = cache.get_cache_region()
    jo = None
    if type(dict.get(domain_uuid)) is NoValue:
        jo = Journal.factory(domain_uuid)
        dict.set(domain_uuid, jo)
    else:
        jo = dict.get(domain_uuid)

    jo.resetState()


#  Qemu guest agent is used to check VM status
#  The checking pre-conditions are as follows:
#  - VM is running
#  - VM has Qemu guest agent installed
#
#  then status is determined by
#  - VM is guest-agent-pingable or not
#
# Note: checkGuests function is called by the scheduler
class QemuGuestAgent(object):

    def __init__(self):
        super(QemuGuestAgent, self).__init__()
        self.notifier = masakari.SendNotification()

    # _thresholdsCrossing
    #
    # We only issue a notification
    # to masakari-engine if the VM is 'already' in the error state
    # when there are consecutive guest_ping failures.
    # Suggested value for guest_monitoring_failure_threshold >= 3
    #
    # Note: When operators are trying to gracefully shutdown VM,
    # QEMU may take time to powering-off.
    # E.g. When you do <nova list> or <openstack server show>
    # you may see that QEMU is active but monitoring may fail
    # due to VM is still "powering-off"
    #  Status  | Task State   | Power State
    #  ACTIVE  | powering-off | Running
    def _thresholdsCrossing(self, domain):
        if (((self.getVmFsm(domain.UUIDString()).current_state) == 'error')
            and
            (self._getJournalObject(domain.UUIDString()).getFailedCount()
                > ICONF.guest_monitoring_failure_threshold)):
            LOG.debug('_thresholdsCrossing:' + domain.UUIDString())
            LOG.debug(self._getJournalObject(
                domain.UUIDString()).getFailedCount())
            return True
        else:
            return False

    def _masakari_notifier(self, domain_uuid):
        if self._getJournalObject(domain_uuid).getSentNotification():
            LOG.debug('notifier.send_notification Skipped:' + domain_uuid)
        else:
            hostname = socket.gethostname()
            noticeType = ec.EventConstants.TYPE_VM
            current_time = timeutils.utcnow()
            event = {
                'notification': {
                    'type': noticeType,
                    'hostname': hostname,
                    'generated_time': current_time,
                    'payload': {
                        'event': 'QEMU_GUEST_AGENT_ERROR',
                        'instance_uuid': domain_uuid,
                        'vir_domain_event': 'STOPPED_FAILED'
                    }
                }
            }
            try:
                self.notifier.send_notification(CONF.callback.retry_max,
                                        CONF.callback.retry_interval,
                                        event)
                self._getJournalObject(domain_uuid).processEvent('report')
                self._getJournalObject(domain_uuid).setSentNotification(True)
            except Exception:
                LOG.warning('Failed to send notification for instance '
                            '%s in %s.', domain_uuid, get_function_name(),
                            exc_info=True)

    def _qemuAgentGuestPing(self, domain, timeout, flags=0):
        def _no_heartbeat(domain_uuid):
            # Also advance the FSM
            self.getVmFsm(domain_uuid).processEvent('guest_not_pingable')
            self._getJournalObject(domain_uuid).incrementFailedCount()

        def _with_heartbeat(domain_uuid):
            #The order matters as we want to decrease the counter first
            self._getJournalObject(domain_uuid).resetFailedCount()
            self.getVmFsm(domain_uuid).processEvent('guest_pingable')

        def _record(result):
            if result is None:
                LOG.debug(domain.UUIDString()
                    + '\tqemu-ga_guest-ping is not responding.')

                if self._thresholdsCrossing(domain):
                    self._masakari_notifier(domain.UUIDString())

                _no_heartbeat(domain.UUIDString())
            else:
                _with_heartbeat(domain.UUIDString())

        """Send a Guest Agent ping to domain """
        # must pass domain._o to the c library as virDomainPtr
        ret = libvirtmod_qemu.virDomainQemuAgentCommand(domain._o,
            '{"execute": "guest-ping"}', timeout, flags)

        _record(ret)

    def _getJournalObject(self, domain_uuid):
        """Function: To get the dictionary

        :param domain: QEMU domain
        :return: the journal object referred by domain_uuid
        """

        dict = cache.get_cache_region()
        if type(dict.get(domain_uuid)) is NoValue:
            jo = Journal.factory(domain_uuid)
            dict.set(domain_uuid, jo)
            return jo
        else:
            return dict.get(domain_uuid)

    def getVmFsm(self, domain_uuid):
        """Function: To get the VM Finite State Machine from
        the dictionary

        :param domain: QEMU domain
        :return: FSM object
        """
        dict = cache.get_cache_region()

        if type(dict.get(domain_uuid)) is NoValue:
            jo = Journal.factory(domain_uuid)
            dict.set(domain_uuid, jo)
            return jo
        else:
            return dict.get(domain_uuid)

    def _hasQemuGuestAgent(self, domain):
        """Function: To check whether the VM has an QEMU Guest Agent
        by examining the qemu.guest_agent sock

        First check if libvirt is running or not, then sock

        :param domain: QEMU domain
        :return: true or false
        """

        def qemuGuestAgentPathMatch(path):
            SOCK = ICONF.qemu_guest_agent_sock_path
            return re.match('%s' % SOCK, path)

        state, reason = domain.state()
        # First check if libvirt is running or not
        if state != libvirt.VIR_DOMAIN_RUNNING:
            return False

        xmlDesc = domain.XMLDesc()
        tree = etree.fromstring(xmlDesc)
        ''' Example
        <channel type='unix'>
         <source mode='bind' path=
          '/var/lib/libvirt/qemu/org.qemu.guest_agent.0.instance-00000004.sock'/>
         <target type=
          'virtio' name='org.qemu.guest_agent.0' state='connected'/>
         <alias name='channel0'/>
         <address type='virtio-serial' controller='0' bus='0' port='1'/>
        </channel>
        '''
        try:
            source = tree.find("devices/channel/source")
            if (source is not None):
                mode = source.get('mode')
                path = source.get('path')
                # There should be a bind for a sock file for qemu guest_agent
                if (qemuGuestAgentPathMatch(path) and mode == 'bind'):
                    return True
        except Exception:
            LOG.warning("Failed to check QEMU guest agent presence.",
            exc_info=True)

        return False

    def checkGuests(self):
        """Function: Check QEMU Guests

        Condition: VM under intrusive monitoring must have QEMU agent client
        configured, installed and qemu "guest-agent-pingable".
        """

        try:
            conn = libvirt.open(CONF.libvirt.connection_uri)

            for domain_id in conn.listDomainsID():
                try:
                    domain = conn.lookupByID(domain_id)
                    if self._hasQemuGuestAgent(domain):
                        @utils.synchronized(domain.UUIDString())
                        def do_qemuAgentGuestPing(domain, timeout):
                            self._qemuAgentGuestPing(domain, timeout)
                        do_qemuAgentGuestPing(domain,
                            ICONF.guest_monitoring_timeout)
                except libvirt.libvirtError as le:
                    LOG.warning(le)
                    continue
        except Exception:
            LOG.warning("Exception during guest checks.",
                        exc_info=True)


def reschedule(action, sleep_time=1):
    """Eventlet Sleep for the specified number of seconds.

    :param sleep_time: seconds to sleep; if None, no sleep;
    """
    LOG.debug('At reschedule')
    if sleep_time is not None:
        LOG.debug('Action %s sleep for %s seconds' % (
            action.id, sleep_time))
        eventlet.sleep(sleep_time)


def sleep(sleep_time):
    """Interface for sleeping."""

    eventlet.sleep(sleep_time)



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generic Node base class for all workers that run on hosts."""

import os
import sys

from oslo_log import log as logging
from oslo_service import service
from oslo_utils import importutils

import masakarimonitors.conf
from masakarimonitors.i18n import _
from masakarimonitors import utils


LOG = logging.getLogger(__name__)

CONF = masakarimonitors.conf.CONF


class Service(service.Service):
    """Service object for binaries running on hosts.

       A service takes a manager.
    """

    def __init__(self, host, binary, manager):
        super(Service, self).__init__()
        self.host = host
        self.binary = binary
        self.manager_class_name = manager
        manager_class = importutils.import_class(self.manager_class_name)
        self.manager = manager_class(host=self.host)

    def __repr__(self):
        return "<%(cls_name)s: host=%(host)s, binary=%(binary)s, " \
               "manager_class_name=%(manager)s>" %\
               {
                   'cls_name': self.__class__.__name__,
                   'host': self.host,
                   'binary': self.binary,
                   'manager': self.manager_class_name
               }

    def start(self):
        LOG.info('Starting %s', self.binary)
        self.basic_config_check()
        self.manager.init_host()
        self.manager.main()

    def __getattr__(self, key):
        manager = self.__dict__.get('manager', None)
        return getattr(manager, key)

    @classmethod
    def create(cls, host=None, binary=None, manager=None):
        """Instantiates class and passes back application object.

        :param host: defaults to CONF.hostname
        :param binary: defaults to basename of executable
        :param manager: defaults to CONF.<Latter part of binary>_manager

        """
        if not host:
            host = CONF.hostname
        if not binary:
            binary = os.path.basename(sys.argv[0])

        if not manager:
            manager_cls = ('%s_manager' %
                           binary.rpartition('masakarimonitors-')[2])
            manager = CONF.get(manager_cls, None)

        service_obj = cls(host, binary, manager)

        return service_obj

    def kill(self):
        """Destroy the service object in the datastore.

        NOTE: Although this method is not used anywhere else than tests, it is
        convenient to have it here, so the tests might easily and in clean way
        stop and remove the service_ref.

        """
        self.stop()

    def stop(self):
        LOG.info('Stopping %s', self.binary)
        self.manager.stop()
        super(Service, self).stop()

    def basic_config_check(self):
        """Perform basic config checks before starting processing."""
        # Make sure the tempdir exists and is writable
        try:
            with utils.tempdir():
                pass
        except Exception as e:
            LOG.error('Temporary directory is invalid: %s', e)
            sys.exit(1)

    def reset(self):
        self.manager.reset()


# NOTE: the global launcher is to maintain the existing
#       functionality of calling service.serve +
#       service.wait
_launcher = None


def serve(server, workers=None):
    global _launcher
    if _launcher:
        raise RuntimeError(_('serve() can only be called once'))

    _launcher = service.launch(CONF, server, workers=workers)


def wait():
    _launcher.wait()



#    Copyright 2017 NTT Data.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import textwrap
from unittest import mock

import pycodestyle
import testtools

from masakarimonitors.hacking import checks


class HackingTestCase(testtools.TestCase):
    """This class tests the hacking checks in masakarimonitors.hacking.checks
    by passing strings to the check methods like the pycodestyle/flake8 parser
    would.

    The parser loops over each line in the file and then passes the
    parameters to the check method. The parameter names in the check method
    dictate what type of object is passed to the check method.

    The parameter types are::

        logical_line: A processed line with the following modifications:
            - Multi-line statements converted to a single line.
            - Stripped left and right.
            - Contents of strings replaced with "xxx" of same length.
            - Comments removed.
        physical_line: Raw line of text from the input file.
        lines: a list of the raw lines from the input file
        tokens: the tokens that contribute to this logical line
        line_number: line number in the input file
        total_lines: number of lines in the input file
        blank_lines: blank lines before this one
        indent_char: indentation character in this file (" " or "\t")
        indent_level: indentation (with tabs expanded to multiples of 8)
        previous_indent_level: indentation on previous line
        previous_logical: previous logical line
        filename: Path of the file being run through pycodestyle

    When running a test on a check method the return will be False/None if
    there is no violation in the sample input. If there is an error a tuple is
    returned with a position in the line, and a message. So to check the result
    just assertTrue if the check is expected to fail and assertFalse if it
    should pass.
    """

    # We are patching pycodestyle so that only the check under test is actually
    # installed.
    @mock.patch('pycodestyle._checks',
                {'physical_line': {}, 'logical_line': {}, 'tree': {}})
    def _run_check(self, code, checker, filename=None):
        pycodestyle.register_check(checker)

        lines = textwrap.dedent(code).strip().splitlines(True)

        checker = pycodestyle.Checker(filename=filename, lines=lines)
        checker.check_all()
        checker.report._deferred_print.sort()
        return checker.report._deferred_print

    def _assert_has_errors(self, code, checker, expected_errors=None,
                           filename=None):
        actual_errors = [e[:3] for e in
                         self._run_check(code, checker, filename)]
        self.assertEqual(expected_errors or [], actual_errors)

    def _assert_has_no_errors(self, code, checker, filename=None):
        self._assert_has_errors(code, checker, filename=filename)

    def test_check_explicit_underscore_import(self):
        self.assertEqual(len(list(checks.check_explicit_underscore_import(
            "LOG.info(_('My info message'))",
            "masakarimonitors/tests/other_files.py"))), 1)
        self.assertEqual(len(list(checks.check_explicit_underscore_import(
            "msg = _('My message')",
            "masakarimonitors/tests/other_files.py"))), 1)
        self.assertEqual(len(list(checks.check_explicit_underscore_import(
            "from masakarimonitors.i18n import _",
            "masakarimonitors/tests/other_files.py"))), 0)
        self.assertEqual(len(list(checks.check_explicit_underscore_import(
            "LOG.info(_('My info message'))",
            "masakarimonitors/tests/other_files.py"))), 0)
        self.assertEqual(len(list(checks.check_explicit_underscore_import(
            "msg = _('My message')",
            "masakarimonitors/tests/other_files.py"))), 0)
        self.assertEqual(len(list(checks.check_explicit_underscore_import(
            "from masakarimonitors.i18n import _",
            "masakarimonitors/tests/other_files2.py"))), 0)
        self.assertEqual(len(list(checks.check_explicit_underscore_import(
            "msg = _('My message')",
            "masakarimonitors/tests/other_files2.py"))), 0)
        self.assertEqual(len(list(checks.check_explicit_underscore_import(
            "_ = translations.ugettext",
            "masakarimonitors/tests/other_files3.py"))), 0)
        self.assertEqual(len(list(checks.check_explicit_underscore_import(
            "msg = _('My message')",
            "masakarimonitors/tests/other_files3.py"))), 0)

    def test_no_translate_logs(self):
        self.assertEqual(1, len(list(checks.no_translate_logs(
            "LOG.error(_LE('foo'))"))))

        self.assertEqual(0, len(list(checks.no_translate_logs(
            "LOG.debug('foo')"))))

        self.assertEqual(1, len(list(checks.no_translate_logs(
            "LOG.info(_LI('foo'))"))))

        self.assertEqual(1, len(list(checks.no_translate_logs(
            "LOG.warning(_LW('foo'))"))))

        self.assertEqual(1, len(list(checks.no_translate_logs(
            "LOG.critical(_LC('foo'))"))))

    def test_no_log_warn(self):
        code = """
                  LOG.warn("LOG.warn is deprecated")
               """
        errors = [(1, 0, 'M305')]
        self._assert_has_errors(code, checks.no_log_warn,
                                expected_errors=errors)
        code = """
                  LOG.warning("LOG.warn is deprecated")
               """
        self._assert_has_no_errors(code, checks.no_log_warn)



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Starter script for Masakari Host Monitor."""

import sys

from oslo_log import log as logging

import masakarimonitors.conf
from masakarimonitors import config
from masakarimonitors import service
from masakarimonitors import utils


CONF = masakarimonitors.conf.CONF


def main():
    config.parse_args(sys.argv)
    logging.setup(CONF, "masakarimonitors")
    utils.monkey_patch()

    server = service.Service.create(binary='masakarimonitors-hostmonitor')
    service.serve(server)
    service.wait()



# Copyright(c) 2024 Samsung SDS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import deque

import eventlet

from kubernetes import client
from kubernetes import config

from oslo_log import log
from oslo_utils import timeutils

import socket

from masakarimonitors import conf
from masakarimonitors.ha import masakari
from masakarimonitors.hostmonitor import driver
from masakarimonitors.objects import event_constants as ec

LOG = log.getLogger(__name__)
CONF = conf.CONF


class KubernetesCheck(driver.DriverBase):
    """Check host status by k8s"""

    def __init__(self):
        super(KubernetesCheck, self).__init__()
        self.hostname = socket.gethostname()
        self.monitoring_interval = CONF.host.monitoring_interval
        self.monitoring_samples = CONF.host.monitoring_samples
        self.notifier = masakari.SendNotification()
        self.monitoring_data = {}
        self.last_status = {}
        self.running = False

        config.load_incluster_config()
        self.kube_client = client.CoreV1Api()

    def update_monitoring_data(self):
        """update monitoring data from k8s"""
        LOG.debug("update monitoring data from k8s.")
        node_list = self.kube_client.list_node(
            label_selector=CONF.kubernetes.monitoring_node_labels
        ).items

        # Remove non-exist node from monitoring_data
        self.monitoring_data = {
            node: status for node, status in self.monitoring_data.items()
            if node in [i.metadata.name for i in node_list]
        }

        for host in node_list:
            hostname = host.metadata.name

            for condition in host.status.conditions:
                if condition.type != "Ready":
                    continue

                host_health_history = self.monitoring_data.get(
                    hostname, deque([], maxlen=self.monitoring_samples))
                host_health_history.appendleft(condition.status)
                LOG.debug("hostname: %s, condition: %s",
                          hostname, host_health_history)

                # Status is 'True' if node is healthy else 'False' or 'unknown'
                # Most up-to-date state is stored in index 0
                self.monitoring_data[hostname] = host_health_history

    @staticmethod
    def _event(host, host_health):
        host_status = ec.EventConstants.HOST_STATUS_NORMAL
        LOG.info("Host %s needs recovery, health status: %s." %
                 (host, host_health))

        event = {
            'notification': {
                'type': ec.EventConstants.TYPE_COMPUTE_HOST,
                'hostname': host,
                'generated_time': timeutils.utcnow(),
                'payload': {
                    'event': ec.EventConstants.EVENT_STOPPED,
                    'host_status': host_status
                }
            }
        }

        return event

    def poll_hosts(self):
        """poll and check hosts health"""
        for host, health_history in self.monitoring_data.items():
            if not self.last_status:
                self.last_status[host] = self.monitoring_data[host][0]

            if host == self.hostname:
                continue

            if len(health_history) < self.monitoring_samples:
                continue

            is_host_alive = "True" in health_history

            if not is_host_alive and self.last_status.get(host):
                event = self._event(host, is_host_alive)
                self.notifier.send_notification(
                    CONF.host.api_retry_max,
                    CONF.host.api_retry_interval,
                    event)

            self.last_status[host] = is_host_alive

    def stop(self):
        self.running = False

    def monitor_hosts(self):
        self.running = True
        while self.running:
            try:
                self.update_monitoring_data()
                LOG.info("Monitoring data: %s", self.monitoring_data)
                self.poll_hosts()
            except Exception as e:
                LOG.exception("Exception when host-monitor by k8s: %s", e)

            eventlet.greenthread.sleep(self.monitoring_interval)



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class EventConstants(object):
    # Define event types.
    TYPE_PROCESS = "PROCESS"
    TYPE_COMPUTE_HOST = "COMPUTE_HOST"
    TYPE_VM = "VM"

    # Define event details.
    EVENT_STARTED = "STARTED"
    EVENT_STOPPED = "STOPPED"

    # Define host status.
    HOST_STATUS_NORMAL = "NORMAL"
    HOST_STATUS_UNKNOWN = "UNKNOWN"

    # Define cluster status.
    CLUSTER_STATUS_ONLINE = "ONLINE"
    CLUSTER_STATUS_OFFLINE = "OFFLINE"



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import testtools
from unittest import mock
from xml.etree import ElementTree

import eventlet

from masakarimonitors.hostmonitor.host_handler import parse_cib_xml

eventlet.monkey_patch(os=False)

CIB_XML = '<cib have-quorum="1">' \
          '  <configuration>' \
          '    <crm_config>test</crm_config>' \
          '    <nodes>' \
          '      <node id="1084754452" uname="masakari-node"/>' \
          '      <node id="1084754453" uname="compute-node"/>' \
          '    </nodes>' \
          '    <resources>' \
          '      <group id="grpStonith1">' \
          '        <primitive id="stnt11" type="external/stonith-helper">' \
          '          <instance_attributes id="stnt11-instance_attributes">' \
          '            <nvpair name="hostlist" value="masakari-node"/>' \
          '          </instance_attributes>' \
          '          <operations>' \
          '            <op name="start"/>' \
          '            <op name="monitor"/>' \
          '            <op name="stop"/>' \
          '          </operations>' \
          '        </primitive>' \
          '        <primitive id="stnt12" type="external/ipmi">' \
          '          <instance_attributes id="stnt12-instance_attributes">' \
          '            <nvpair name="pcmk_reboot_timeout" value="60s"/>' \
          '            <nvpair name="hostname" value="masakari-node"/>' \
          '            <nvpair name="ipaddr" value="192.168.10.20"/>' \
          '            <nvpair name="userid" value="admin"/>' \
          '            <nvpair name="passwd" value="password"/>' \
          '            <nvpair name="interface" value="lanplus"/>' \
          '          </instance_attributes>' \
          '          <operations>' \
          '            <op name="start"/>' \
          '            <op name="monitor"/>' \
          '            <op name="stop"/>' \
          '          </operations>' \
          '        </primitive>' \
          '      </group>' \
          '      <group id="grpStonith2">' \
          '        <primitive id="stnt21" type="external/stonith-helper">' \
          '          <instance_attributes id="stnt21-instance_attributes">' \
          '            <nvpair name="hostlist" value="compute-node"/>' \
          '          </instance_attributes>' \
          '          <operations>' \
          '            <op name="start"/>' \
          '            <op name="monitor"/>' \
          '            <op name="stop"/>' \
          '          </operations>' \
          '        </primitive>' \
          '        <primitive id="stnt22" type="external/ipmi">' \
          '          <instance_attributes id="stnt22-instance_attributes">' \
          '            <nvpair name="pcmk_reboot_timeout" value="60s"/>' \
          '            <nvpair name="hostname" value="compute-node"/>' \
          '            <nvpair name="ipaddr" value="192.168.10.21"/>' \
          '            <nvpair name="userid" value="admin"/>' \
          '            <nvpair name="passwd" value="password"/>' \
          '            <nvpair name="interface" value="lanplus"/>' \
          '          </instance_attributes>' \
          '          <operations>' \
          '            <op name="start"/>' \
          '            <op name="monitor"/>' \
          '            <op name="stop"/>' \
          '          </operations>' \
          '        </primitive>' \
          '      </group>' \
          '    </resources>' \
          '    <constraints>' \
          '      <rsc_location id="loc_grpStonith1" rsc="grpStonith1">' \
          '        <rule test="foo"/>' \
          '      </rsc_location>' \
          '    </constraints>' \
          '  </configuration>' \
          '  <status>' \
          '    <node_state uname="masakari-node" crmd="online">' \
          '      <test foo="foo"/>' \
          '    </node_state>' \
          '    <node_state crmd="online" uname="compute-node">' \
          '      <test foo="foo"/>' \
          '    </node_state>' \
          '  </status>' \
          '</cib>'
CIB_TAG = ElementTree.fromstring(CIB_XML)


class TestParseCibXml(testtools.TestCase):

    def setUp(self):
        super(TestParseCibXml, self).setUp()

    @mock.patch.object(ElementTree, 'fromstring')
    def test_set_cib_xml(self,
                         mock_fromstring):

        obj = parse_cib_xml.ParseCibXml()
        mock_fromstring.return_value = CIB_TAG
        obj.set_cib_xml(CIB_XML)

    def test_have_quorum(self):

        obj = parse_cib_xml.ParseCibXml()
        obj.set_cib_xml(CIB_XML)
        self.assertEqual(1, obj.have_quorum())

    def test_get_node_state_tag_list(self):

        obj = parse_cib_xml.ParseCibXml()
        obj.set_cib_xml(CIB_XML)

        node_state_tag_list = obj.get_node_state_tag_list()

        for node_state_tag in node_state_tag_list:
            self.assertEqual('online', node_state_tag.get('crmd'))

    def test_get_stonith_ipmi_params(self):

        obj = parse_cib_xml.ParseCibXml()
        obj.set_cib_xml(CIB_XML)

        ipmi_values = obj.get_stonith_ipmi_params('compute-node')

        self.assertEqual('192.168.10.21', ipmi_values['ipaddr'])
        self.assertEqual('admin', ipmi_values['userid'])
        self.assertEqual('password', ipmi_values['passwd'])
        self.assertEqual('lanplus', ipmi_values['interface'])



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import socket
import testtools
from unittest import mock

import ddt
import eventlet
from oslo_utils import timeutils

import masakarimonitors.conf
from masakarimonitors.ha import masakari
from masakarimonitors.objects import event_constants as ec
from masakarimonitors.processmonitor.process_handler import handle_process
from masakarimonitors import utils

CONF = masakarimonitors.conf.CONF
eventlet.monkey_patch(os=False)

NOVA_COMPUTE = 'nova-compute'

MOCK_PROCESS_LIST = [
    {
        'id': 1,
        'process_name': '/usr/local/bin/nova-compute',
        'start_command': 'mock_start_command',
        'pre_start_command': 'mock_pre_start_command',
        'post_start_command': 'mock_post_start_command',
        'restart_command': 'mock_restart_command',
        'pre_restart_command': 'mock_pre_restart_command',
        'post_restart_command': 'mock_post_restart_command',
        'run_as_root': True
    },
]

MOCK_DOWN_PROCESS_LIST = [
    {
        'id': 1,
        'process_name': '/usr/local/bin/nova-compute',
        'start_command': 'mock_start_command',
        'pre_start_command': 'mock_pre_start_command',
        'post_start_command': 'mock_post_start_command',
        'restart_command': 'mock_restart_command',
        'pre_restart_command': 'mock_pre_restart_command',
        'post_restart_command': 'mock_post_restart_command',
        'run_as_root': True
    },
]

PS_RESULT = \
    "\n" \
    "UID  PID   PPID C STIME TTY TIME     CMD\n" \
    "root 11187 1    0 18:52 ?   00:00:00 /usr/local/bin/nova-compute\n"


@ddt.ddt
class TestHandleProcess(testtools.TestCase):

    def setUp(self):
        super(TestHandleProcess, self).setUp()

    def test_set_process_list(self):
        process_list = MOCK_PROCESS_LIST
        obj = handle_process.HandleProcess()
        obj.set_process_list(process_list)

        self.assertEqual(process_list, obj.process_list)

    @ddt.data("/usr/local/bin/nova-compute", "nova-compute")
    def test_make_event(self, process_name):
        obj = handle_process.HandleProcess()
        event = obj._make_event(process_name)
        self.assertEqual(
            NOVA_COMPUTE,
            event['notification']['payload']['process_name'])

    @mock.patch.object(utils, 'execute')
    def test_start_processes(self,
                             mock_execute):
        process_list = MOCK_PROCESS_LIST
        obj = handle_process.HandleProcess()
        obj.set_process_list(process_list)

        mock_execute.side_effect = [('test_stdout', ''),
                                    ('test_stdout', ''),
                                    ('test_stdout', 'test_stderr')]

        obj.start_processes()

        mock_execute.assert_any_call(
            MOCK_PROCESS_LIST[0].get('pre_start_command'),
            run_as_root=MOCK_PROCESS_LIST[0].get('run_as_root'))
        mock_execute.assert_any_call(
            MOCK_PROCESS_LIST[0].get('start_command'),
            run_as_root=MOCK_PROCESS_LIST[0].get('run_as_root'))
        mock_execute.assert_any_call(
            MOCK_PROCESS_LIST[0].get('post_start_command'),
            run_as_root=MOCK_PROCESS_LIST[0].get('run_as_root'))

    @mock.patch.object(utils, 'execute')
    def test_start_processes_pre_cmd_fail(self, mock_execute):
        process_list = MOCK_PROCESS_LIST
        obj = handle_process.HandleProcess()
        obj.set_process_list(process_list)

        mock_execute.return_value = ('test_stdout', 'test_stderr')

        obj.start_processes()

        mock_execute.assert_called_once_with(
            MOCK_PROCESS_LIST[0].get('pre_start_command'),
            run_as_root=MOCK_PROCESS_LIST[0].get('run_as_root'))

    @mock.patch.object(utils, 'execute')
    def test_monitor_processes(self,
                               mock_execute):
        process_list = MOCK_PROCESS_LIST
        obj = handle_process.HandleProcess()
        obj.set_process_list(process_list)

        mock_execute.return_value = (PS_RESULT, '')

        down_process_list = obj.monitor_processes()
        self.assertEqual([], down_process_list)
        mock_execute.assert_called_once_with(
            'ps', '-ef', run_as_root=False)

    @mock.patch.object(utils, 'execute')
    def test_monitor_processes_not_found(self, mock_execute):
        process_list = MOCK_PROCESS_LIST
        obj = handle_process.HandleProcess()
        obj.set_process_list(process_list)

        mock_execute.return_value = ('', '')

        down_process_list = obj.monitor_processes()
        self.assertEqual(MOCK_PROCESS_LIST, down_process_list)
        mock_execute.assert_called_once_with(
            'ps', '-ef', run_as_root=False)

    @mock.patch.object(utils, 'execute')
    def test_monitor_processes_exception(self, mock_execute):
        process_list = MOCK_PROCESS_LIST
        obj = handle_process.HandleProcess()
        obj.set_process_list(process_list)

        mock_execute.side_effect = Exception("Test exception.")

        down_process_list = obj.monitor_processes()
        self.assertEqual([], down_process_list)

    @mock.patch.object(utils, 'execute')
    def test_restart_processes(self,
                               mock_execute):
        process_list = MOCK_PROCESS_LIST
        obj = handle_process.HandleProcess()
        obj.set_process_list(process_list)
        down_process_list = MOCK_DOWN_PROCESS_LIST

        mock_execute.side_effect = [('test_stdout', ''),
                                    ('test_stdout', ''),
                                    ('test_stdout', '')]

        obj.restart_processes(down_process_list)

        mock_execute.assert_any_call(
            down_process_list[0].get('pre_restart_command'),
            run_as_root=down_process_list[0].get('run_as_root'))
        mock_execute.assert_any_call(
            down_process_list[0].get('restart_command'),
            run_as_root=down_process_list[0].get('run_as_root'))
        mock_execute.assert_any_call(
            down_process_list[0].get('post_restart_command'),
            run_as_root=down_process_list[0].get('run_as_root'))
        self.assertEqual([], obj.restart_failure_list)

    @mock.patch.object(utils, 'execute')
    def test_restart_processes_failed_to_restart_previously(
        self, mock_execute):
        process_list = MOCK_PROCESS_LIST
        obj = handle_process.HandleProcess()
        obj.set_process_list(process_list)
        restart_failure_list = [MOCK_DOWN_PROCESS_LIST[0].get('process_name')]
        obj.restart_failure_list = restart_failure_list
        down_process_list = MOCK_DOWN_PROCESS_LIST

        obj.restart_processes(down_process_list)

        self.assertEqual(restart_failure_list, obj.restart_failure_list)
        mock_execute.assert_not_called()

    @mock.patch.object(masakari.SendNotification, 'send_notification')
    @mock.patch.object(timeutils, 'utcnow')
    @mock.patch.object(eventlet.greenthread, 'sleep')
    @mock.patch.object(utils, 'execute')
    def test_restart_processes_pre_restart_command_retry_over(
        self, mock_execute, mock_sleep, mock_utcnow, mock_send_notification):
        process_list = MOCK_PROCESS_LIST
        obj = handle_process.HandleProcess()
        obj.set_process_list(process_list)
        down_process_list = MOCK_DOWN_PROCESS_LIST

        mock_execute.side_effect = [('test_stdout', 'test_stderr'),
                                    ('test_stdout', 'test_stderr'),
                                    ('test_stdout', 'test_stderr'),
                                    ('test_stdout', 'test_stderr')]
        mock_sleep.return_value = None
        current_time = timeutils.utcnow()
        mock_utcnow.return_value = current_time
        mock_send_notification.return_value = None

        obj.restart_processes(down_process_list)

        pre_execute_count = CONF.process.restart_retries + 1
        self.assertEqual(pre_execute_count, mock_execute.call_count)

        for var in range(0, mock_execute.call_count):
            args, kwargs = mock_execute.call_args_list[var]
            self.assertEqual(
                (down_process_list[0].get('pre_restart_command'),),
                args)
            self.assertEqual({'run_as_root': True}, kwargs)

        event = {
            'notification': {
                'type': ec.EventConstants.TYPE_PROCESS,
                'hostname': socket.gethostname(),
                'generated_time': current_time,
                'payload': {
                    'event': ec.EventConstants.EVENT_STOPPED,
                    'process_name': NOVA_COMPUTE
                }
            }
        }
        mock_send_notification.assert_called_once_with(
            CONF.process.api_retry_max,
            CONF.process.api_retry_interval,
            event)

        self.assertEqual(
            [down_process_list[0].get('process_name')],
            obj.restart_failure_list)

    @mock.patch.object(masakari.SendNotification, 'send_notification')
    @mock.patch.object(timeutils, 'utcnow')
    @mock.patch.object(eventlet.greenthread, 'sleep')
    @mock.patch.object(utils, 'execute')
    def test_restart_processes_restart_command_retry_over(
        self, mock_execute, mock_sleep, mock_utcnow, mock_send_notification):
        process_list = MOCK_PROCESS_LIST
        obj = handle_process.HandleProcess()
        obj.set_process_list(process_list)
        down_process_list = MOCK_DOWN_PROCESS_LIST

        mock_execute.side_effect = [('test_stdout', ''),
                                    ('test_stdout', 'test_stderr'),
                                    ('test_stdout', ''),
                                    ('test_stdout', 'test_stderr'),
                                    ('test_stdout', ''),
                                    ('test_stdout', 'test_stderr'),
                                    ('test_stdout', ''),
                                    ('test_stdout', 'test_stderr')]
        mock_sleep.return_value = None
        current_time = timeutils.utcnow()
        mock_utcnow.return_value = current_time
        mock_send_notification.return_value = None

        obj.restart_processes(down_process_list)

        pre_execute_count = CONF.process.restart_retries + 1
        execute_count = CONF.process.restart_retries + 1
        total_execute_count = pre_execute_count + execute_count
        self.assertEqual(total_execute_count, mock_execute.call_count)

        for var in range(0, mock_execute.call_count):
            # Execute order of restart_command is the second.
            execute_order = 2

            if (var + 1) % execute_order == 0:
                args, kwargs = mock_execute.call_args_list[var]
                self.assertEqual(
                    (down_process_list[0].get('restart_command'),),
                    args)
                self.assertEqual({'run_as_root': True}, kwargs)

        event = {
            'notification': {
                'type': ec.EventConstants.TYPE_PROCESS,
                'hostname': socket.gethostname(),
                'generated_time': current_time,
                'payload': {
                    'event': ec.EventConstants.EVENT_STOPPED,
                    'process_name': NOVA_COMPUTE
                }
            }
        }
        mock_send_notification.assert_called_once_with(
            CONF.process.api_retry_max,
            CONF.process.api_retry_interval,
            event)
        self.assertEqual(
            [down_process_list[0].get('process_name')],
            obj.restart_failure_list)

    @mock.patch.object(masakari.SendNotification, 'send_notification')
    @mock.patch.object(timeutils, 'utcnow')
    @mock.patch.object(eventlet.greenthread, 'sleep')
    @mock.patch.object(utils, 'execute')
    def test_restart_processes_post_restart_command_retry_over(
        self, mock_execute, mock_sleep, mock_utcnow, mock_send_notification):
        process_list = MOCK_PROCESS_LIST
        obj = handle_process.HandleProcess()
        obj.set_process_list(process_list)
        down_process_list = MOCK_DOWN_PROCESS_LIST

        mock_execute.side_effect = [('test_stdout', ''),
                                    ('test_stdout', ''),
                                    ('test_stdout', 'test_stderr'),
                                    ('test_stdout', ''),
                                    ('test_stdout', ''),
                                    ('test_stdout', 'test_stderr'),
                                    ('test_stdout', ''),
                                    ('test_stdout', ''),
                                    ('test_stdout', 'test_stderr'),
                                    ('test_stdout', ''),
                                    ('test_stdout', ''),
                                    ('test_stdout', 'test_stderr')]
        mock_sleep.return_value = None
        current_time = timeutils.utcnow()
        mock_utcnow.return_value = current_time
        mock_send_notification.return_value = None

        obj.restart_processes(down_process_list)

        pre_execute_count = CONF.process.restart_retries + 1
        execute_count = CONF.process.restart_retries + 1
        post_execute_count = CONF.process.restart_retries + 1
        total_execute_count = \
            pre_execute_count + execute_count + post_execute_count
        self.assertEqual(total_execute_count, mock_execute.call_count)

        for var in range(0, mock_execute.call_count):
            # Execute order of restart_command is the third.
            execute_order = 3

            if (var + 1) % execute_order == 0:
                args, kwargs = mock_execute.call_args_list[var]
                self.assertEqual(
                    (down_process_list[0].get('post_restart_command'),),
                    args)
                self.assertEqual({'run_as_root': True}, kwargs)

        event = {
            'notification': {
                'type': ec.EventConstants.TYPE_PROCESS,
                'hostname': socket.gethostname(),
                'generated_time': current_time,
                'payload': {
                    'event': ec.EventConstants.EVENT_STOPPED,
                    'process_name': NOVA_COMPUTE
                }
            }
        }
        mock_send_notification.assert_called_once_with(
            CONF.process.api_retry_max,
            CONF.process.api_retry_interval,
            event)
        self.assertEqual(
            [down_process_list[0].get('process_name')],
            obj.restart_failure_list)

    @mock.patch.object(utils, 'execute')
    def test_execute_cmd_exception(self, mock_execute):
        mock_execute.side_effect = Exception("Test exception.")

        obj = handle_process.HandleProcess()
        ret = obj._execute_cmd(MOCK_PROCESS_LIST[0].get('start_command'),
                               MOCK_PROCESS_LIST[0].get('run_as_root'))

        self.assertEqual(ret, 1)



# Copyright (c) 2017, NTT Data
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re

from hacking import core

"""
Guidelines for writing new hacking checks

 - Use only for masakarimonitors specific tests. OpenStack general tests
   should be submitted to the common 'hacking' module.
 - Pick numbers in the range M3xx. Find the current test with
   the highest allocated number and then pick the next value.
 - Keep the test method code in the source file ordered based
   on the M3xx value.
 - List the new rule in the top level HACKING.rst file
 - Add test cases for each new rule to masakarimonitors/tests/unit/
   test_hacking.py

"""

UNDERSCORE_IMPORT_FILES = []

translated_log = re.compile(
    r"(.)*LOG\.(audit|error|info|critical|exception)"
    r"\(\s*_\(\s*('|\")")
string_translation = re.compile(r"[^_]*_\(\s*('|\")")
underscore_import_check = re.compile(r"(.)*import _(.)*")
underscore_import_check_multi = re.compile(r"(.)*i18n\s+import(.)* _, (.)*")
# We need this for cases where they have created their own _ function.
custom_underscore_check = re.compile(r"(.)*_\s*=\s*(.)*")
log_translation = re.compile(
    r"(.)*LOG\."
    r"(audit|debug|error|info|warn|warning|critical|exception)"
    r"\("
    r"(_|_LC|_LE|_LI|_LW)"
    r"\(")


@core.flake8ext
def check_explicit_underscore_import(logical_line, filename):
    """Check for explicit import of the _ function

    We need to ensure that any files that are using the _() function
    to translate logs are explicitly importing the _ function.  We
    can't trust unit test to catch whether the import has been
    added so we need to check for it here.
    """

    # Build a list of the files that have _ imported.  No further
    # checking needed once it is found.
    for file in UNDERSCORE_IMPORT_FILES:
        if file in filename:
            return
    if (underscore_import_check.match(logical_line) or
            underscore_import_check_multi.match(logical_line) or
            custom_underscore_check.match(logical_line)):
        UNDERSCORE_IMPORT_FILES.append(filename)
    elif (translated_log.match(logical_line) or
         string_translation.match(logical_line)):
        yield (0, "M301: Found use of _() without explicit import of _ !")


@core.flake8ext
def no_translate_logs(logical_line):
    """Check that logging doesn't translate messages
    M302
    """
    if log_translation.match(logical_line):
        yield (0, "M302 Don't translate log messages!")


@core.flake8ext
def no_log_warn(logical_line):
    """Disallow 'LOG.warn('

    LOG.warn() is deprecated and LOG.warning should be used instead.
    https://docs.python.org/3/library/logging.html#logging.warning

    N352
    """

    msg = ("M305: LOG.warn is deprecated, please use LOG.warning!")
    if "LOG.warn(" in logical_line:
        yield (0, msg)



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import socket
import testtools
import threading
from unittest import mock
import uuid

import eventlet
from oslo_utils import excutils
from oslo_utils import timeutils

from masakarimonitors.instancemonitor.libvirt_handler import callback
from masakarimonitors.instancemonitor.libvirt_handler import eventfilter
from masakarimonitors.instancemonitor.libvirt_handler \
    import eventfilter_table as evft
from masakarimonitors.objects import event_constants as ec

eventlet.monkey_patch(os=False)


class TestEventFilter(testtools.TestCase):

    def setUp(self):
        super(TestEventFilter, self).setUp()

    @mock.patch.object(excutils, 'save_and_reraise_exception')
    @mock.patch.object(callback.Callback, 'libvirt_event_callback')
    @mock.patch.object(timeutils, 'utcnow')
    def test_vir_event_filter(self, mock_utcnow, mock_libvirt_event_callback,
        mock_save_and_reraise_exception):

        current_time = timeutils.utcnow()
        mock_utcnow.return_value = current_time
        mock_libvirt_event_callback.return_value = None
        mock_save_and_reraise_exception.return_value = None

        obj = eventfilter.EventFilter()
        eventID = 0
        eventType = 5
        detail = 5
        uuID = uuid.uuid4()
        obj.vir_event_filter(eventID, eventType, detail, uuID)

        mock_libvirt_event_callback.assert_called_once_with(
            evft.eventID_dic[eventID],
            evft.detail_dic[eventID][eventType][detail],
            uuID,
            ec.EventConstants.TYPE_VM,
            socket.gethostname(),
            current_time)
        mock_save_and_reraise_exception.assert_not_called()

    @mock.patch.object(excutils, 'save_and_reraise_exception')
    @mock.patch.object(callback.Callback, 'libvirt_event_callback')
    def test_vir_event_filter_unmatched(self, mock_libvirt_event_callback,
        mock_save_and_reraise_exception):

        mock_libvirt_event_callback.return_value = None
        mock_save_and_reraise_exception.return_value = None

        obj = eventfilter.EventFilter()
        eventID = 0
        eventType = 5
        detail = 2
        uuID = uuid.uuid4()
        obj.vir_event_filter(eventID, eventType, detail, uuID)

        mock_libvirt_event_callback.assert_not_called()
        mock_save_and_reraise_exception.assert_not_called()

    @mock.patch.object(excutils, 'save_and_reraise_exception')
    @mock.patch.object(callback.Callback, 'libvirt_event_callback')
    def test_vir_event_filter_key_error(self, mock_libvirt_event_callback,
        mock_save_and_reraise_exception):

        mock_libvirt_event_callback.return_value = None
        mock_save_and_reraise_exception.return_value = None

        obj = eventfilter.EventFilter()
        eventID = 0
        eventType = 0
        detail = 0
        uuID = uuid.uuid4()
        obj.vir_event_filter(eventID, eventType, detail, uuID)

        mock_libvirt_event_callback.assert_not_called()
        mock_save_and_reraise_exception.assert_not_called()

    @mock.patch.object(excutils, 'save_and_reraise_exception')
    @mock.patch.object(callback.Callback, 'libvirt_event_callback')
    @mock.patch.object(threading, 'Thread')
    def test_vir_event_filter_type_error(self, mock_Thread,
        mock_libvirt_event_callback, mock_save_and_reraise_exception):

        mock_Thread.side_effect = TypeError("Threading exception.")
        mock_libvirt_event_callback.return_value = None
        mock_save_and_reraise_exception.return_value = None

        obj = eventfilter.EventFilter()
        eventID = 0
        eventType = 5
        detail = 5
        uuID = uuid.uuid4()
        obj.vir_event_filter(eventID, eventType, detail, uuID)

        mock_libvirt_event_callback.assert_not_called()
        mock_save_and_reraise_exception.assert_not_called()

    @mock.patch.object(excutils, 'save_and_reraise_exception')
    @mock.patch.object(callback.Callback, 'libvirt_event_callback')
    @mock.patch.object(threading, 'Thread')
    def test_vir_event_filter_index_error(self, mock_Thread,
        mock_libvirt_event_callback, mock_save_and_reraise_exception):

        mock_Thread.side_effect = IndexError("Threading exception.")
        mock_libvirt_event_callback.return_value = None
        mock_save_and_reraise_exception.return_value = None

        obj = eventfilter.EventFilter()
        eventID = 0
        eventType = 5
        detail = 5
        uuID = uuid.uuid4()
        obj.vir_event_filter(eventID, eventType, detail, uuID)

        mock_libvirt_event_callback.assert_not_called()
        mock_save_and_reraise_exception.assert_not_called()

    @mock.patch.object(callback.Callback, 'libvirt_event_callback')
    @mock.patch.object(threading, 'Thread')
    def test_vir_event_filter_other_exception(self, mock_Thread,
        mock_libvirt_event_callback):

        mock_Thread.side_effect = NameError("Threading exception.")
        mock_libvirt_event_callback.return_value = None

        obj = eventfilter.EventFilter()
        eventID = 0
        eventType = 5
        detail = 5
        uuID = uuid.uuid4()
        self.assertRaises(NameError, obj.vir_event_filter, eventID, eventType,
            detail, uuID)

        mock_libvirt_event_callback.assert_not_called()



# Copyright(c) 2018 WindRiver Systems
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import eventlet
import libvirt
import testtools
from unittest import mock

from masakarimonitors.introspectiveinstancemonitor import instance

eventlet.monkey_patch(os=False)


class TestMonitorManager(testtools.TestCase):

    def setUp(self):
        super(TestMonitorManager, self).setUp()

    @mock.patch.object(libvirt, 'virEventRunDefaultImpl')
    def test_vir_event_loop_native_run(self, mock_virEventRunDefaultImpl):
        mock_virEventRunDefaultImpl.side_effect = Exception("Test exception.")

        obj = instance.IntrospectiveInstanceMonitorManager()
        exception_flag = False
        try:
            obj._vir_event_loop_native_run()
        except Exception:
            exception_flag = True

        self.assertTrue(exception_flag)
        mock_virEventRunDefaultImpl.assert_called_once()



# Copyright(c) 2019 Canonical Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from xml.etree import ElementTree

from oslo_log import log as oslo_logging

LOG = oslo_logging.getLogger(__name__)


class ParseCrmMonXml(object):
    """ParseCrmMonXml class

    This class parses the crmmon xml.
    """

    def __init__(self):
        self.crmmon_tag = None

    def set_crmmon_xml(self, crmmon_xml):
        """Set xml.etree.ElementTree.Element object.

        This method receives string of crmmon xml, and convert it
        to xml.etree.ElementTree.Element object.

        :params crmmon_xml: String of crmmon xml
        """
        # Convert xml.etree.ElementTree.Element object.
        self.crmmon_tag = ElementTree.fromstring(crmmon_xml)

    def has_quorum(self):
        """Answers if cluster has quorum or not.

        :returns: True if cluster has quorum, False otherwise.
        """
        current_dc = self._get_current_dc()
        return current_dc.get('with_quorum') == 'true'

    def _get_summary(self):
        child_list = list(self.crmmon_tag)
        for child in child_list:
            if child.tag == 'summary':
                return child
        return None

    def _get_current_dc(self):
        child_list = list(self._get_summary())
        for child in child_list:
            if child.tag == 'current_dc':
                return child
        return None

    def _get_nodes(self):
        # status tag exists in the crmmon tag.
        if self.crmmon_tag is None:
            return None
        child_list = list(self.crmmon_tag)
        for child in child_list:
            if child.tag == 'nodes':
                return child
        return None

    def _get_node_states(self, nodes_tag):
        node_state_tag_list = []

        # node_state tag exists in the status tag.
        child_list = list(nodes_tag)
        for child in child_list:
            if child.tag == 'node':
                node_state_tag_list.append(child)

        return node_state_tag_list

    def get_node_state_tag_list(self):
        """Get node_state tag list.

        This method gets node_state tag list from crmmon xml.

        :returns: node_state tag list
        """
        # Get status tag.
        nodes_tag = self._get_nodes()
        if nodes_tag is None:
            LOG.error("crm_mon xml doesn't have nodes tag.")
            return []

        # Get node_state tag list.
        node_state_tag_list = self._get_node_states(nodes_tag)
        if len(node_state_tag_list) == 0:
            LOG.error("crm_mon xml doesn't have online tag.")

        return node_state_tag_list



# Copyright(c) 2019 Inspur
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg


consul_opts = [
    cfg.StrOpt('agent_manage',
               help='Addr for local consul agent in management datacenter.'),
    cfg.StrOpt('agent_tenant',
               help='Addr for local consul agent in tenant datacenter.'),
    cfg.StrOpt('agent_storage',
               help='Addr for local consul agent in storage datacenter.'),
    cfg.StrOpt('matrix_config_file',
               help='Config file for consul health action matrix.'),
]


def register_opts(conf):
    conf.register_opts(consul_opts, group='consul')


def list_opts():
    return {
        'consul': consul_opts
    }



# Copyright(c) 2019 Inspur
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import eventlet
import socket

from collections import deque
from oslo_log import log
from oslo_utils import timeutils

from masakarimonitors import conf
from masakarimonitors.ha import masakari
from masakarimonitors.hostmonitor.consul_check import consul_helper
from masakarimonitors.hostmonitor.consul_check import matrix_helper
from masakarimonitors.hostmonitor import driver
from masakarimonitors.objects import event_constants as ec

LOG = log.getLogger(__name__)
CONF = conf.CONF


class ConsulCheck(driver.DriverBase):
    """Check host status by consul"""

    def __init__(self):
        super(ConsulCheck, self).__init__()
        self.hostname = socket.gethostname()
        self.monitoring_interval = CONF.host.monitoring_interval
        self.monitoring_samples = CONF.host.monitoring_samples
        self.matrix_manager = matrix_helper.MatrixManager(CONF)
        self.consul_manager = consul_helper.ConsulManager(CONF)
        self.notifier = masakari.SendNotification()
        self._matrix = None
        self._sequence = None
        self.monitoring_data = {}
        self.last_host_health = {}

    @property
    def matrix(self):
        if not self._matrix:
            self._matrix = self.matrix_manager.get_matrix()
        return self._matrix

    @property
    def sequence(self):
        if not self._sequence:
            self._sequence = self.matrix_manager.get_sequence()
        return self._sequence

    def _formate_health(self, host_health):
        formate_health = {}
        for i in range(len(host_health)):
            layer = "%s-interface" % self.sequence[i]
            state = host_health[i]
            formate_health[layer] = state

        return formate_health

    def _event(self, host, host_health):
        host_status = ec.EventConstants.HOST_STATUS_NORMAL
        if 'down' not in host_health:
            event_type = ec.EventConstants.EVENT_STARTED
            cluster_status = ec.EventConstants.CLUSTER_STATUS_ONLINE
        else:
            actions = self.get_action_from_matrix(host_health)
            if 'recovery' in actions:
                LOG.info("Host %s needs recovery, health status: %s." %
                         (host, str(self._formate_health(host_health))))
                event_type = ec.EventConstants.EVENT_STOPPED
                cluster_status = ec.EventConstants.CLUSTER_STATUS_OFFLINE
            else:
                return None

        # TODO(suzhengwei): Add host status detail in the payload to show
        # host status in all Consul clusters.
        event = {
            'notification': {
                'type': ec.EventConstants.TYPE_COMPUTE_HOST,
                'hostname': host,
                'generated_time': timeutils.utcnow(),
                'payload': {
                    'event': event_type,
                    'cluster_status': cluster_status,
                    'host_status': host_status
                }
            }
        }

        return event

    def update_monitoring_data(self):
        '''update monitoring data from consul clusters'''
        LOG.debug("update monitoring data from consul.")
        # Get current host health in sequence [x, y, z]. The example of
        # the return value is {'node01':[x, y, z], 'node02':[x, y, z]...}
        cluster_health = self.consul_manager.get_health(self.sequence)
        LOG.debug("Current cluster state: %s.", cluster_health)
        # Reassemble host health history with the latest host health.
        # Example of 'host_health_history' is [[x, y, x], [x, y, z]...]
        for host, health in cluster_health.items():
            host_health_history = self.monitoring_data.get(
                host, deque([], maxlen=self.monitoring_samples))
            host_health_history.append(health)
            self.monitoring_data[host] = host_health_history

    def get_host_health(self, host):
        health_history = self.monitoring_data.get(host, [])
        if len(health_history) < self.monitoring_samples:
            LOG.debug("Not enough monitoring data for host %s", host)
            return None

        # Caculate host health from host health history.
        # Only continous 'down' represents the interface 'down',
        # while continous 'up' represents the interface 'up'.
        host_sequence_health = []
        host_health_history = list(zip(*health_history))
        for i in range(0, len(host_health_history)):
            if ('up' in host_health_history[i] and
                    'down' in host_health_history[i]):
                host_sequence_health.append(None)
            else:
                host_sequence_health.append(host_health_history[i][0])

        return host_sequence_health

    def _host_health_changed(self, host, health):
        last_health = self.last_host_health.get(host)
        if last_health is None:
            self.last_host_health[host] = health
            return False

        if health != last_health:
            self.last_host_health[host] = health
            return True
        else:
            return False

    def get_action_from_matrix(self, host_health):
        for health_action in self.matrix:
            matrix_health = health_action["health"]
            matrix_action = health_action["action"]
            if host_health == matrix_health:
                return matrix_action

        return []

    def poll_hosts(self):
        '''poll and check hosts health'''
        for host in self.monitoring_data.keys():

            if host == self.hostname:
                continue

            host_health = self.get_host_health(host)
            if host_health is None:
                continue

            if not self._host_health_changed(host, host_health):
                continue

            # it will send notifition to trigger host failure recovery
            # according to defined HA strategy
            event = self._event(host, host_health)
            if event:
                self.notifier.send_notification(
                    CONF.host.api_retry_max,
                    CONF.host.api_retry_interval,
                    event)

    def stop(self):
        self.running = False

    def monitor_hosts(self):
        self.running = True
        while self.running:
            try:
                self.update_monitoring_data()
                self.poll_hosts()
            except Exception as e:
                LOG.exception("Exception when host-monitor by consul: %s", e)

            eventlet.greenthread.sleep(self.monitoring_interval)



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import yaml

import eventlet
from oslo_log import log as oslo_logging

import masakarimonitors.conf
from masakarimonitors import manager
from masakarimonitors.processmonitor.process_handler import handle_process

LOG = oslo_logging.getLogger(__name__)
CONF = masakarimonitors.conf.CONF


class ProcessmonitorManager(manager.Manager):
    """Manages the masakari-processmonitor."""

    def __init__(self, *args, **kwargs):
        super(ProcessmonitorManager, self).__init__(
            service_name="processmonitor", *args, **kwargs)
        self.process_handler = handle_process.HandleProcess()

    def _load_process_list(self):
        try:
            process_list = yaml.safe_load(open(CONF.process.process_list_path))
            LOG.debug("Loaded process list. %s" % process_list)

            return process_list
        except yaml.YAMLError as e:
            LOG.exception("YAMLError caught: %s", e)
            return
        except Exception as e:
            LOG.exception("Exception caught: %s", e)
            return

    def stop(self):
        self.running = False

    def main(self):
        """Main method."""

        try:
            # Load process list.
            process_list = self._load_process_list()
            if process_list is None:
                LOG.error("Failed to load process list file.")
                return

            # Set process_list object to the process handler.
            self.process_handler.set_process_list(process_list)

            # Initial start of processes.
            self.process_handler.start_processes()

            self.running = True
            while self.running:
                # Monitor processes.
                down_process_list = self.process_handler.monitor_processes()

                if len(down_process_list) != 0:
                    # Restart down processes.
                    self.process_handler.restart_processes(down_process_list)
                else:
                    # Since no down process, clear the restart_failure_list
                    self.process_handler.restart_failure_list[:] = []

                # Reload process list and set to the process handler.
                process_list = self._load_process_list()
                if process_list is None:
                    LOG.error("Failed to reload process list file.")
                    break
                self.process_handler.set_process_list(process_list)

                eventlet.greenthread.sleep(CONF.process.check_interval)

        except Exception as e:
            LOG.exception("Exception caught: %s", e)
            return

        return



# Copyright(c) 2019 Canonical Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import testtools

from masakarimonitors.hostmonitor.host_handler import parse_crmmon_xml


CRMMON_XML = '<?xml version="1.0"?>' \
             '<crm_mon version="1.1.18">' \
             '    <summary>' \
             '        <stack type="corosync" />' \
             '        <current_dc present="true" with_quorum="true" />' \
             '    </summary>' \
             '    <nodes>' \
             '        <node name="node-1" id="1001" online="true" />' \
             '        <node name="node-2" id="1002" online="false" />' \
             '        <node name="node-3" id="1003" online="true" />' \
             '    </nodes>' \
             '</crm_mon>'

CRMMON_XML_NO_QUORUM = \
    '<?xml version="1.0"?>' \
    '<crm_mon version="1.1.18">' \
    '    <summary>' \
    '        <stack type="corosync" />' \
    '        <current_dc present="true" with_quorum="false" />' \
    '    </summary>' \
    '</crm_mon>'

CRMMON_NONODES_XML = '<?xml version="1.0"?>' \
                     '<crm_mon version="1.1.18">' \
                     '    <nodes>' \
                     '    </nodes>' \
                     '</crm_mon>'

CRMMON_NONODES_TAG_XML = '<?xml version="1.0"?>' \
                         '<crm_mon version="1.1.18">' \
                         '</crm_mon>'


class TestParseCrmMonXml(testtools.TestCase):

    def setUp(self):
        super(TestParseCrmMonXml, self).setUp()

    def test_set_crmmon_xml(self):
        obj = parse_crmmon_xml.ParseCrmMonXml()
        obj.set_crmmon_xml(CRMMON_XML)

    def test_has_quorum(self):
        obj = parse_crmmon_xml.ParseCrmMonXml()
        obj.set_crmmon_xml(CRMMON_XML)
        self.assertEqual(True, obj.has_quorum())

    def test_has_quorum_no_quorum(self):
        obj = parse_crmmon_xml.ParseCrmMonXml()
        obj.set_crmmon_xml(CRMMON_XML_NO_QUORUM)
        self.assertEqual(False, obj.has_quorum())

    def test_get_node_state_tag_list(self):
        obj = parse_crmmon_xml.ParseCrmMonXml()
        obj.set_crmmon_xml(CRMMON_XML)

        node_state_tag_list = obj.get_node_state_tag_list()

        expected = {
            'node-1': 'true',
            'node-2': 'false',
            'node-3': 'true'}

        for node_state_tag in node_state_tag_list:
            self.assertEqual(
                expected[node_state_tag.get('name')],
                node_state_tag.get('online'))

    def test_get_node_state_tag_list_unset(self):
        obj = parse_crmmon_xml.ParseCrmMonXml()
        self.assertEqual(obj.get_node_state_tag_list(), [])

    def test_get_node_state_tag_list_nonodes(self):
        obj = parse_crmmon_xml.ParseCrmMonXml()
        obj.set_crmmon_xml(CRMMON_NONODES_XML)
        self.assertEqual(obj.get_node_state_tag_list(), [])

    def test_get_node_state_tag_list_nonodes_tag(self):
        obj = parse_crmmon_xml.ParseCrmMonXml()
        obj.set_crmmon_xml(CRMMON_NONODES_TAG_XML)
        self.assertEqual(obj.get_node_state_tag_list(), [])



# Copyright(c) 2024 Samsung SDS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import deque
from unittest import mock

from kubernetes import client
from kubernetes.client import V1Node
from kubernetes.client import V1NodeCondition
from kubernetes.client import V1NodeList
from kubernetes.client import V1NodeStatus
from kubernetes.client import V1ObjectMeta
from kubernetes import config

from oslo_config import fixture as fixture_config

import testtools

import masakarimonitors.conf
from masakarimonitors.ha import masakari
from masakarimonitors.hostmonitor.kubernetes_check import manager

CONF = masakarimonitors.conf.CONF


class TestKubernetesCheck(testtools.TestCase):

    @mock.patch.object(config, 'load_incluster_config')
    @mock.patch.object(client, 'CoreV1Api')
    def setUp(self, mock_core_v1_api, mock_load_incluster_config):
        super(TestKubernetesCheck, self).setUp()
        self.CONF = self.useFixture(fixture_config.Config()).conf
        self.CONF.kubernetes.monitoring_node_labels = \
            'openstack-compute-node=enabled'
        self.host_monitor = manager.KubernetesCheck()
        self.host_monitor.monitoring_samples = 1
        self.host_monitor.monitoring_data = {}
        self.last_status = {}

    def test_update_monitoring_data(self):
        mock_list_node = V1NodeList(
            items=[
                V1Node(
                    metadata=V1ObjectMeta(
                        name="node0" + str(num),
                        labels={
                            "openstack-compute-node": "enabled"
                        }
                    ),
                    status=V1NodeStatus(conditions=[
                        V1NodeCondition(status="False", type="MemoryPressure"),
                        V1NodeCondition(status="False", type="DiskPressure"),
                        V1NodeCondition(status="False", type="PIDPressure"),
                        V1NodeCondition(status="True", type="Ready")
                    ])
                )
                for num in range(1, 4)
            ]
        )

        with mock.patch.object(self.host_monitor.kube_client, 'list_node',
                               return_value=mock_list_node):
            self.host_monitor.update_monitoring_data()
            excepted_monitoring_data = {
                "node01": deque(['True'],
                                maxlen=self.host_monitor.monitoring_samples),
                "node02": deque(['True'],
                                maxlen=self.host_monitor.monitoring_samples),
                "node03": deque(['True'],
                                maxlen=self.host_monitor.monitoring_samples)
            }
        self.assertEqual(excepted_monitoring_data,
                         self.host_monitor.monitoring_data)

    def test_update_monitoring_data_exclude_node(self):

        mock_list_node = V1NodeList(
            items=[
                V1Node(
                    metadata=V1ObjectMeta(
                        name="node0" + str(num),
                        labels={
                            "openstack-compute-node": "enabled",
                            "monitoring": "enabled"
                        }
                    ),
                    status=V1NodeStatus(conditions=[
                        V1NodeCondition(status="False", type="MemoryPressure"),
                        V1NodeCondition(status="False", type="DiskPressure"),
                        V1NodeCondition(status="False", type="PIDPressure"),
                        V1NodeCondition(status="True", type="Ready")
                    ])
                )
                for num in range(1, 4)
            ]
        )

        with mock.patch.object(self.host_monitor.kube_client, 'list_node',
                               return_value=mock_list_node):
            self.host_monitor.update_monitoring_data()

            mock_list_node.items.pop()

            self.host_monitor.update_monitoring_data()
            excepted_monitoring_data = {
                "node01": deque(['True', 'True'],
                                maxlen=self.host_monitor.monitoring_samples),
                "node02": deque(['True', 'True'],
                                maxlen=self.host_monitor.monitoring_samples)
            }

        self.assertEqual(2, len(self.host_monitor.monitoring_data))
        self.assertEqual(excepted_monitoring_data,
                         self.host_monitor.monitoring_data)

    @mock.patch.object(masakari.SendNotification, 'send_notification')
    def test_poll_hosts_with_healthy_nodes(self, mock_send_notification):
        self.host_monitor.monitoring_data = {
            "node01": deque(['True'],
                            maxlen=self.host_monitor.monitoring_samples),
            "node02": deque(['True'],
                            maxlen=self.host_monitor.monitoring_samples),
            "node03": deque(['True'],
                            maxlen=self.host_monitor.monitoring_samples),
        }

        self.host_monitor.poll_hosts()
        mock_send_notification.assert_not_called()

    @mock.patch.object(masakari.SendNotification, 'send_notification')
    def test_poll_hosts_with_healthy_nodes_3samples(
            self, mock_send_notification):
        self.host_monitor.monitoring_samples = 3
        self.host_monitor.monitoring_data = {
            "node01": deque(['True', 'True', 'True'],
                            maxlen=self.host_monitor.monitoring_samples),
            "node02": deque(['Unknown', 'True', 'True'],
                            maxlen=self.host_monitor.monitoring_samples),
            "node03": deque(['True', 'Unknown', 'Unknown'],
                            maxlen=self.host_monitor.monitoring_samples),
        }

        self.host_monitor.poll_hosts()
        mock_send_notification.assert_not_called()

    @mock.patch.object(masakari.SendNotification, 'send_notification')
    @mock.patch.object(manager.KubernetesCheck, '_event')
    def test_poll_hosts_with_unhealthy_node(self, mock_event,
                                            mock_send_notification):
        self.host_monitor.last_status = {
            "node01": True,
            "node02": True,
            "node03": True
        }
        self.host_monitor.monitoring_samples = 3
        self.host_monitor.monitoring_data = {
            "node01": deque(['True', 'True', 'True'],
                            maxlen=self.host_monitor.monitoring_samples),
            "node02": deque(['Unknown', 'Unknown', 'Unknown'],
                            maxlen=self.host_monitor.monitoring_samples),
            "node03": deque(['Unknown', 'True', 'Unknown'],
                            maxlen=self.host_monitor.monitoring_samples),
        }

        test_event = {'notification': 'test'}
        mock_event.return_value = test_event

        self.host_monitor.poll_hosts()
        mock_send_notification.assert_called_once_with(
            CONF.host.api_retry_max, CONF.host.api_retry_interval, test_event)

    @mock.patch.object(masakari.SendNotification, 'send_notification')
    def test_poll_hosts_with_having_less_healthy_history_than_samples(
            self, mock_send_notification):
        self.host_monitor.monitoring_samples = 3
        self.host_monitor.monitoring_data = {
            "node01": deque(['True', 'True'],
                            maxlen=self.host_monitor.monitoring_samples),
            "node02": deque(['True', 'Unknown'],
                            maxlen=self.host_monitor.monitoring_samples),
            "node03": deque(['Unknown', 'Unknown'],
                            maxlen=self.host_monitor.monitoring_samples),
        }

        self.host_monitor.poll_hosts()
        mock_send_notification.assert_not_called()



# Copyright(c) 2018 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import libvirt
import testtools
from unittest import mock
import uuid


from masakarimonitors.introspectiveinstancemonitor import instance as object
from masakarimonitors.introspectiveinstancemonitor import qemu_utils
from masakarimonitors.introspectiveinstancemonitor import scheduler


class TestQemuUtils(testtools.TestCase):

    def setup(self):
        super(TestQemuUtils, self).setUp()

    @mock.patch.object(qemu_utils.libvirt, 'virDomain')
    def test_getVmFsm(self, mock_domain):
        """To test the state machines

        Initial stage should be dicovery and will advance to a
        healthy stage after enough pingable events. Also,
        it should reach an error stage from healthy after
        a pingable event failure.
        """

        reference = qemu_utils.QemuGuestAgent()
        mock_domain.UUID.return_value = uuid.uuid4()
        reference.getVmFsm(mock_domain)
        self.assertEqual(reference.getVmFsm(mock_domain).current_state,
            'discovery')
        reference.getVmFsm(mock_domain).process_event('guest_not_pingable')
        self.assertEqual(reference.getVmFsm(mock_domain).current_state,
            'discovery')
        reference.getVmFsm(mock_domain).process_event('guest_pingable')
        self.assertEqual(reference.getVmFsm(mock_domain).current_state,
            'healthy')
        reference.getVmFsm(mock_domain).process_event('guest_pingable')
        self.assertEqual(reference.getVmFsm(mock_domain).current_state,
            'healthy')
        reference.getVmFsm(mock_domain).process_event('guest_not_pingable')
        self.assertEqual(reference.getVmFsm(mock_domain).current_state,
            'error')

    @mock.patch.object(qemu_utils.libvirt, 'virDomain')
    def test_hasQemuGuestAgent(self, mock_domain):
        mock_domain.UUID.return_value = 'testuuid'
        mock_domain.state.return_value = libvirt.VIR_DOMAIN_RUNNING, 'reason'
        mock_domain.XMLDesc.return_value = """<domain>
  <devices>
    <interface type="network">
      <target dev="vnet0"/>
    </interface>
    <interface type="network">
      <target dev="vnet1"/>
    </interface>
  </devices>
</domain>"""
        obj = qemu_utils.QemuGuestAgent()
        self.assertFalse(obj._hasQemuGuestAgent(mock_domain))
        mock_domain.XMLDesc.return_value = """<domain>
  <devices>
    <interface type="network">
      <target dev="vnet0"/>
    </interface>
    <interface type="network">
      <target dev="vnet1"/>
    </interface>
    <channel type='unix'>
      <source
        mode='bind'
        path=
        '/var/lib/libvirt/qemu/org.qemu.guest_agent.0.instance-00000003.sock'/>
      <target
        type='virtio'
        name='org.qemu.guest_agent.0' state='disconnected'/>
      <alias name='channel0'/>
      <address type='virtio-serial' controller='0' bus='0' port='1'/>
    </channel>
  </devices>
</domain>"""
        obj = qemu_utils.QemuGuestAgent()
        self.assertTrue(obj._hasQemuGuestAgent(mock_domain))

    @mock.patch.object(scheduler.ThreadGroupManager, '_service_task')
    @mock.patch.object(qemu_utils, 'resetJournal')
    def test_resetJournal(self, mock_resetJournal, mock_service_task):

        mock_resetJournal.return_value = None

        obj = object.IntrospectiveInstanceMonitorManager()

        event_id = 0
        domain_uuid = uuid.uuid4()
        event_type = libvirt.VIR_DOMAIN_EVENT_STARTED
        detail = libvirt.VIR_DOMAIN_EVENT_STARTED_BOOTED

        obj._reset_journal(event_id, event_type, detail, domain_uuid)

        mock_resetJournal.assert_called_once_with(domain_uuid)



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import socket

import eventlet
from oslo_log import log as oslo_logging
from oslo_utils import timeutils

import masakarimonitors.conf
from masakarimonitors.ha import masakari
from masakarimonitors.objects import event_constants as ec
from masakarimonitors import utils

LOG = oslo_logging.getLogger(__name__)
CONF = masakarimonitors.conf.CONF


class HandleProcess(object):
    """Handle process."""

    def __init__(self):
        self.process_list = None
        self.restart_failure_list = []
        self.notifier = masakari.SendNotification()

    def set_process_list(self, process_list):
        """Set process list object.

        :param process_list: process list object
        """
        self.process_list = process_list

    def _execute_cmd(self, cmd_str, run_as_root):

        # Split command string and delete empty elements.
        command = cmd_str.split()
        command = filter(lambda x: x != '', command)

        try:
            # Execute start command.
            out, err = utils.execute(*command,
                                     run_as_root=run_as_root)

            if out:
                msg = ("CMD '%s' output stdout: %s") % (cmd_str, out)
                LOG.info("%s", msg)

            if err:
                msg = ("CMD '%s' output stderr: %s") % (cmd_str, err)
                LOG.warning("%s", msg)
                return 1

        except Exception as e:
            msg = ("CMD '%s' raised exception: %s") % (cmd_str, e)
            LOG.error("%s", e)
            return 1

        return 0

    def start_processes(self):
        """Initial start of processes.

        This method starts the processes using start command written in the
        process list.
        """
        for process in self.process_list:
            cmd_str = process['start_command']
            pre_cmd_str = process['pre_start_command']
            post_cmd_str = process['post_start_command']

            # Execute pre start command.
            if pre_cmd_str:
                ret = self._execute_cmd(pre_cmd_str, process['run_as_root'])
                if ret != 0:
                    continue

            # Execute start command.
            LOG.info("Start of process with executing command: %s", cmd_str)
            self._execute_cmd(cmd_str, process['run_as_root'])

            # Execute post start command.
            if post_cmd_str:
                ret = self._execute_cmd(post_cmd_str, process['run_as_root'])

    def monitor_processes(self):
        """Monitor processes.

        This method monitors the processes using process name written in the
        process list.

        :returns: List of down process
        """
        down_process_list = []
        for process in self.process_list:
            process_name = process['process_name']

            try:
                # Execute monitoring command.
                out, err = utils.execute('ps', '-ef', run_as_root=False)
                if process_name in out:
                    LOG.debug("Process '%s' is found." % process_name)
                else:
                    # Append down_process_list.
                    down_process_list.append(process)
                    LOG.warning("Process '%s' is not found.", process_name)
            except Exception as e:
                LOG.error("Monitoring command raised exception: %s", e)

        return down_process_list

    def _make_event(self, process_name):

        hostname = socket.gethostname()
        current_time = timeutils.utcnow()
        event = {
            'notification': {
                'type': ec.EventConstants.TYPE_PROCESS,
                'hostname': hostname,
                'generated_time': current_time,
                'payload': {
                    'event': ec.EventConstants.EVENT_STOPPED,
                    'process_name': process_name.split('/')[-1]
                }
            }
        }

        return event

    def restart_processes(self, down_process_list):
        """Restart processes.

        This method restarts the processes using restart command written in
        the process list.

        :param down_process_list: down process list object
        """
        tmp_restart_failure_list = []
        for down_process in down_process_list:
            # The process which failed to restart previously doesn't restart.
            if down_process['process_name'] in self.restart_failure_list:
                msg = "Process '%s' doesn't be restarted because it failed" \
                      " to restart previously." % down_process['process_name']
                LOG.warning("%s", msg)
                tmp_restart_failure_list.append(down_process['process_name'])
                continue

            cmd_str = down_process['restart_command']
            pre_cmd_str = down_process['pre_restart_command']
            post_cmd_str = down_process['post_restart_command']

            LOG.info("Restart of process with executing command: %s", cmd_str)

            for retries in range(0, CONF.process.restart_retries + 1):

                # Execute pre restart command.
                if pre_cmd_str:
                    ret = self._execute_cmd(pre_cmd_str,
                                            down_process['run_as_root'])
                    if ret != 0:
                        # Failed to restart process.
                        eventlet.greenthread.sleep(
                            CONF.process.restart_interval)
                        continue

                # Execute restart command.
                ret = self._execute_cmd(cmd_str, down_process['run_as_root'])
                if ret != 0:
                    # Failed to restart process.
                    eventlet.greenthread.sleep(CONF.process.restart_interval)
                    continue

                # Execute post restart command.
                if post_cmd_str:
                    ret = self._execute_cmd(post_cmd_str,
                                            down_process['run_as_root'])
                    if ret != 0:
                        # Failed to restart process.
                        eventlet.greenthread.sleep(
                            CONF.process.restart_interval)
                        continue

                # Succeeded in restarting process.
                break

            if retries == CONF.process.restart_retries:
                # Send a notification.
                event = self._make_event(down_process['process_name'])
                self.notifier.send_notification(
                    CONF.process.api_retry_max,
                    CONF.process.api_retry_interval,
                    event)
                # Add the process name which failed to restart to the
                # failure list.
                tmp_restart_failure_list.append(down_process['process_name'])

        # Replace the old restart_failure_list with new one
        self.restart_failure_list = tmp_restart_failure_list[:]



# Copyright(c) 2021 Inspur
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Main abstraction layer for retrieving node status from consul
"""

import consul
from oslo_utils import netutils

from masakarimonitors.i18n import _


DEFAULT_CONSUL_PORT = 8500


class ConsulException(Exception):
    """Base Consul Exception"""
    msg_fmt = _("An unknown exception occurred.")

    def __init__(self, message=None, **kwargs):
        if not message:
            message = self.msg_fmt % kwargs

        super(ConsulException, self).__init__(message)


class ConsulAgentNotExist(ConsulException):
    msg_fmt = _("Consul agent of %(cluster)s not exist.")


class ConsulGetMembersException(ConsulException):
    msg_fmt = _("Failed to get members of %(cluster)s: %(err)s.")


class ConsulManager(object):
    """Consul manager class

    This class helps to pull health data from all consul clusters,
    and return health data in sequence.
    """

    def __init__(self, CONF):
        self.agents = {}
        self.init_agents(CONF)

    def init_agents(self, CONF):
        if CONF.consul.agent_manage:
            addr, port = netutils.parse_host_port(
                CONF.consul.agent_manage, DEFAULT_CONSUL_PORT)
            self.agents['manage'] = ConsulAgent('manage', addr, port)
        if CONF.consul.agent_tenant:
            addr, port = netutils.parse_host_port(
                CONF.consul.agent_tenant, DEFAULT_CONSUL_PORT)
            self.agents['tenant'] = ConsulAgent('tenant', addr, port)
        if CONF.consul.agent_storage:
            addr, port = netutils.parse_host_port(
                CONF.consul.agent_storage, DEFAULT_CONSUL_PORT)
            self.agents['storage'] = ConsulAgent('storage', addr, port)

    def valid_agents(self, sequence):
        for name in sequence:
            if self.agents.get(name) is None:
                raise ConsulAgentNotExist(cluster=name)

    def get_health(self, sequence):
        hosts_health = {}
        all_agents = []
        for name in sequence:
            if name in self.agents:
                consul_agent = self.agents.get(name)
                agent_health = consul_agent.get_health()
                hosts_health[name] = agent_health
                if not all_agents:
                    all_agents = agent_health.keys()

        sequence_hosts_health = {}
        for host in all_agents:
            sequence_hosts_health[host] = []
            for name in sequence:
                if name in self.agents:
                    state = hosts_health[name].get(host)
                    if state:
                        sequence_hosts_health[host].append(state)
                    else:
                        continue

        return sequence_hosts_health


class ConsulAgent(object):
    """Agent to consul cluster"""

    def __init__(self, name, addr=None, port=None):
        self.name = name
        self.addr = addr
        self.port = port
        # connection to consul cluster
        self.cluster = consul.Consul(host=addr, port=self.port)

    def get_agents(self):
        try:
            members = self.cluster.agent.members()
        except Exception as e:
            raise ConsulGetMembersException(cluster=self.name, err=str(e))

        return members

    def get_health(self):
        agents_health = {}
        agents = self.get_agents()
        for agent in agents:
            host = agent.get('Name')
            status = agent.get('Status')
            if status == 1:
                agents_health[host] = 'up'
            else:
                agents_health[host] = 'down'

        return agents_health



# Copyright(c) 2021 Inspur
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import yaml


DEFAULT_SEQUENCE = ['manage', 'tenant', 'storage']

# matrix is combined by health and actions.
# health: [x, x, x], repreasents status of DEFAULT_SEQUENCE.
# action, means which actions it will trigge if host health turns into.
# action choice: 'recovery'.
# 'recovery' means it will trigge one host recovery event.
DEFAULT_MATRIX = [
    {"health": ["up", "up", "up"],
     "action": []},
    {"health": ["up", "up", "down"],
     "action": ["recovery"]},
    {"health": ["up", "down", "up"],
     "action": []},
    {"health": ["up", "down", "down"],
     "action": ["recovery"]},
    {"health": ["down", "up", "up"],
     "action": []},
    {"health": ["down", "up", "down"],
     "action": ["recovery"]},
    {"health": ["down", "down", "up"],
     "action": []},
    {"health": ["down", "down", "down"],
     "action": ["recovery"]},
]


class MatrixManager(object):
    """Matrix Manager"""

    def __init__(self, CONF):
        cfg_file = CONF.consul.matrix_config_file
        matrix_conf = self.load_config(cfg_file)
        if not matrix_conf:
            self.sequence = DEFAULT_SEQUENCE
            self.matrix = DEFAULT_MATRIX
        else:
            self.sequence = matrix_conf.get("sequence")
            self.matrix = matrix_conf.get("matrix")

            self.valid_matrix(self.matrix, self.sequence)

    def load_config(self, cfg_file):
        if not cfg_file or not os.path.exists(cfg_file):
            return None

        with open(cfg_file) as f:
            data = f.read()
        matrix_conf = yaml.safe_load(data)
        return matrix_conf

    def get_sequence(self):
        return self.sequence

    def get_matrix(self):
        return self.matrix

    def valid_matrix(self, matrix, sequence):
        pass



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
import time

import eventlet
import libvirt
from oslo_config import cfg
from oslo_log import log as oslo_logging

from masakarimonitors.instancemonitor.libvirt_handler import eventfilter
from masakarimonitors import manager

LOG = oslo_logging.getLogger(__name__)
CONF = cfg.CONF


class InstancemonitorManager(manager.Manager):
    """Manages the masakari-instancemonitor."""

    def __init__(self, *args, **kwargs):
        super(InstancemonitorManager, self).__init__(
            service_name="instancemonitor", *args, **kwargs)
        self.evf = eventfilter.EventFilter()
        # This keeps track of what thread is running the event loop,
        # (if it is run in a background thread)
        self.event_loop_thread = None

    def _vir_event_loop_native_run(self):
        # Directly run the event loop in the current thread
        while True:
            libvirt.virEventRunDefaultImpl()

    def _vir_event_loop_native_start(self):
        libvirt.virEventRegisterDefaultImpl()
        self.event_loop_thread = threading.Thread(
            target=self._vir_event_loop_native_run,
            name="lib_virt_eventLoop")
        self.event_loop_thread.daemon = True
        self.event_loop_thread.start()

    def _my_domain_event_callback(self, conn, dom, event, detail, opaque):
        self.evf.vir_event_filter(libvirt.VIR_DOMAIN_EVENT_ID_LIFECYCLE,
                                  event, detail, dom.UUIDString())

    def _my_domain_event_reboot_callback(self, conn, dom, opaque):
        self.evf.vir_event_filter(libvirt.VIR_DOMAIN_EVENT_ID_REBOOT,
                                  -1, -1, dom.UUIDString())

    def _my_domain_event_rtc_change_callback(self, conn, dom, utcoffset,
                                             opaque):
        self.evf.vir_event_filter(libvirt.VIR_DOMAIN_EVENT_ID_RTC_CHANGE,
                                  -1, -1, dom.UUIDString())

    def _my_domain_event_watchdog_callback(self, conn, dom, action, opaque):
        self.evf.vir_event_filter(libvirt.VIR_DOMAIN_EVENT_ID_WATCHDOG,
                                  action, -1, dom.UUIDString())

    def _my_domain_event_io_error_callback(self, conn, dom, srcpath,
                                           devalias, action, opaque):
        self.evf.vir_event_filter(libvirt.VIR_DOMAIN_EVENT_ID_IO_ERROR,
                                  action, -1, dom.UUIDString())

    def _my_domain_event_graphics_callback(self, conn, dom, phase, localAddr,
                                           remoteAddr, authScheme, subject,
                                           opaque):
        self.evf.vir_event_filter(libvirt.VIR_DOMAIN_EVENT_ID_GRAPHICS,
                                  -1, phase, dom.UUIDString())

    def _my_domain_event_disk_change_callback(self, conn, dom, oldSrcPath,
                                              newSrcPath, devAlias, reason,
                                              opaque):
        self.evf.vir_event_filter(libvirt.VIR_DOMAIN_EVENT_ID_DISK_CHANGE,
                                  -1, -1, dom.UUIDString())

    def _my_domain_event_io_error_reason_callback(self, conn, dom, srcPath,
                                                  devAlias, action, reason,
                                                  opaque):
        self.evf.vir_event_filter(libvirt.VIR_DOMAIN_EVENT_ID_IO_ERROR_REASON,
                                  -1, -1, dom.UUIDString())

    def _my_domain_event_generic_callback(self, conn, dom, opaque):
        self.evf.vir_event_filter(libvirt.VIR_DOMAIN_EVENT_ID_CONTROL_ERROR,
                                  -1, -1, dom.UUIDString())

    def _err_handler(self, ctxt, err):
        LOG.warning("Error from libvirt : %s", err[2])

    @staticmethod
    def _connect_auth_cb(creds, user_data):
        if len(creds) == 0:
            return 0
        raise Exception("Can not handle authentication request for %d "
                        "credentials" % len(creds))

    def _virt_event(self, uri):
        auth = [[libvirt.VIR_CRED_AUTHNAME,
                 libvirt.VIR_CRED_ECHOPROMPT,
                 libvirt.VIR_CRED_REALM,
                 libvirt.VIR_CRED_PASSPHRASE,
                 libvirt.VIR_CRED_NOECHOPROMPT,
                 libvirt.VIR_CRED_EXTERNAL],
                InstancemonitorManager._connect_auth_cb,
                None]

        flags = libvirt.VIR_CONNECT_RO

        # Run a background thread with the event loop
        self._vir_event_loop_native_start()

        event_callback_handlers = {
            libvirt.VIR_DOMAIN_EVENT_ID_LIFECYCLE:
                self._my_domain_event_callback,
            libvirt.VIR_DOMAIN_EVENT_ID_REBOOT:
                self._my_domain_event_reboot_callback,
            libvirt.VIR_DOMAIN_EVENT_ID_RTC_CHANGE:
                self._my_domain_event_rtc_change_callback,
            libvirt.VIR_DOMAIN_EVENT_ID_IO_ERROR:
                self._my_domain_event_io_error_callback,
            libvirt.VIR_DOMAIN_EVENT_ID_WATCHDOG:
                self._my_domain_event_watchdog_callback,
            libvirt.VIR_DOMAIN_EVENT_ID_GRAPHICS:
                self._my_domain_event_graphics_callback,
            libvirt.VIR_DOMAIN_EVENT_ID_DISK_CHANGE:
                self._my_domain_event_disk_change_callback,
            libvirt.VIR_DOMAIN_EVENT_ID_IO_ERROR_REASON:
                self._my_domain_event_io_error_reason_callback,
            libvirt.VIR_DOMAIN_EVENT_ID_CONTROL_ERROR:
                self._my_domain_event_generic_callback
        }
        # Connect to libvirt - If be disconnected, reprocess.
        self.running = True
        while self.running:
            vc = libvirt.openAuth(uri, auth, flags)

            # Event callback settings
            callback_ids = []
            for event, callback in event_callback_handlers.items():
                cid = vc.domainEventRegisterAny(None, event, callback, None)
                callback_ids.append(cid)

            # Connection monitoring.
            vc.setKeepAlive(5, 3)
            while vc.isAlive() == 1 and self.running:
                eventlet.greenthread.sleep(1)

            # If connection between libvirtd was lost,
            # clear callback connection.
            LOG.warning("Libvirt Connection Closed Unexpectedly.")
            for cid in callback_ids:
                try:
                    vc.domainEventDeregisterAny(cid)
                except Exception:
                    LOG.debug("Failed to deregister libvirt event "
                              "callback %s.", cid, exc_info=True)
            vc.close()
            del vc
            time.sleep(3)

    def stop(self):
        self.running = False

    def main(self):
        """Main method.

        Set the URI, error handler, and executes event loop processing.
        """
        uri = CONF.libvirt.connection_uri
        LOG.debug("Using uri:" + uri)

        # set error handler & do event loop
        libvirt.registerErrorHandler(self._err_handler, '_virt_event')
        self._virt_event(uri)



# Copyright(c) 2018 WindRiver Systems
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg


# Note: this string is being used for regex parsing later with re module.
#
# Use Python's raw string notation for regular expressions and
# uses the backslash character ('\') to indicate special
# forms or to allow special characters to be used without invoking
# their special meaning.
SOCK = r'/var/lib/libvirt/qemu/org\.qemu\.guest_agent\..*\.instance-.*\.sock'

monitor_opts = [
    cfg.IntOpt('guest_monitoring_interval',
               default=10,
               help='''
Guest monitoring interval of VM status (in seconds).
* The value should not be too low as there should not be false negative
* for reporting QEMU_GUEST_AGENT failures
* VM needs time to do powering-off.
* guest_monitoring_interval should be greater than
* the time to SHUTDOWN VM gracefully.
* e.g. | 565da9ba-3c0c-4087-83ca | iim1 | ACTIVE | powering-off | Running
'''),
    cfg.IntOpt('guest_monitoring_timeout',
               default=2,
               help='Guest monitoring timeout (in seconds).'),
    cfg.IntOpt('guest_monitoring_failure_threshold',
               default=3,
               help='Failure threshold before sending notification.'),
    cfg.StrOpt('qemu_guest_agent_sock_path',
               default=SOCK,
               help=r'''
* The file path of qemu guest agent sock.
* Please use Python raw string notation as regular expressions.
  e.g.  r'/var/lib/libvirt/qemu/org\.qemu\.guest_agent\..*\.instance-.*\.sock'
'''),
]


def register_opts(conf):
    conf.register_opts(monitor_opts, group='introspectiveinstancemonitor')


def list_opts():
    return {
        'introspectiveinstancemonitor': monitor_opts
    }



# Copyright(c) 2016 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import socket

from oslo_config import cfg

service_opts = [
    cfg.StrOpt('hostname',
               default=socket.gethostname(),
               deprecated_name="host",
               help='''
Hostname, FQDN or IP address of this host. Must be valid within AMQP key.

Possible values:

* String with hostname, FQDN or IP address. Default is hostname of this host.
'''),
    cfg.StrOpt('instancemonitor_manager',
               default='masakarimonitors.instancemonitor.instance'
                       '.InstancemonitorManager',
               help='Full class name for the Manager for instancemonitor.'),
    cfg.StrOpt('introspectiveinstancemonitor_manager',
               default='masakarimonitors.introspectiveinstancemonitor.instance'
                       '.IntrospectiveInstanceMonitorManager',
               help='Full class name for introspectiveinstancemonitor.'),
    cfg.StrOpt('processmonitor_manager',
               default='masakarimonitors.processmonitor.process'
                       '.ProcessmonitorManager',
               help='Full class name for the Manager for processmonitor.'),
    cfg.StrOpt('hostmonitor_manager',
               default='masakarimonitors.hostmonitor.host'
                       '.HostmonitorManager',
               help='Full class name for the Manager for hostmonitor.'),
    ]


def register_opts(conf):
    conf.register_opts(service_opts)


def list_opts():
    return {'DEFAULT': service_opts}



# Copyright(c) 2017 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import testtools
from unittest import mock
import uuid

import eventlet
from keystoneauth1 import loading as ks_loading
from openstack import connection
from openstack import exceptions
from oslo_utils import timeutils

import masakarimonitors.conf
from masakarimonitors.ha import masakari
from masakarimonitors.objects import event_constants as ec

CONF = masakarimonitors.conf.CONF

PROFILE_TYPE = "ha"
PROFILE_NAME = "masakari"


class FakeResponse(object):

    def __init__(self, status_code=200, headers=None):
        self.status_code = status_code
        self.headers = {
            'content-type': 'application/json',
            'x-openstack-request-id': uuid.uuid4().hex,
        }


class TestSendNotification(testtools.TestCase):

    def setUp(self):
        super(TestSendNotification, self).setUp()
        self.api_retry_max = 3
        self.api_retry_interval = 1
        self.event = {
            'notification': {
                'type': ec.EventConstants.TYPE_COMPUTE_HOST,
                'hostname': 'compute-node1',
                'generated_time': timeutils.utcnow(),
                'payload': {
                    'event': ec.EventConstants.EVENT_STOPPED,
                    'cluster_status': 'OFFLINE',
                    'host_status': ec.EventConstants.HOST_STATUS_NORMAL
                }
            }
        }

    @mock.patch.object(connection, 'Connection')
    @mock.patch.object(ks_loading, 'load_session_from_conf_options')
    @mock.patch.object(ks_loading, 'load_auth_from_conf_options')
    def test_send_notification(
        self, mock_auth, mock_session, mock_connection):

        mock_conn = mock.Mock()
        mock_conn.instance_ha.return_value = mock.Mock()
        mock_conn.instance_ha.create_notification.return_value = mock.Mock()
        mock_connection.return_value = mock_conn

        notifier = masakari.SendNotification()
        notifier.send_notification(
            self.api_retry_max, self.api_retry_interval, self.event)

        mock_conn.instance_ha.create_notification.assert_called_once_with(
            type=self.event['notification']['type'],
            hostname=self.event['notification']['hostname'],
            generated_time=self.event['notification']['generated_time'],
            payload=self.event['notification']['payload'])

        mock_auth.assert_called_once_with(CONF, 'api')
        mock_session.assert_called_once_with(CONF, 'api',
                                             auth=mock_auth.return_value)

    @mock.patch.object(connection, 'Connection')
    @mock.patch.object(ks_loading, 'load_session_from_conf_options')
    @mock.patch.object(ks_loading, 'load_auth_from_conf_options')
    def test_send_notification_409_error(
        self, mock_auth, mock_session, mock_connection):

        mock_conn = mock.Mock()
        mock_conn.instance_ha.return_value = mock.Mock()
        mock_conn.instance_ha.create_notification.return_value = mock.Mock()
        mock_connection.return_value = mock_conn

        response = FakeResponse(status_code=409)
        status_ex = exceptions.HttpException(response=response)

        mock_conn.instance_ha.create_notification.side_effect = status_ex
        notifier = masakari.SendNotification()
        notifier.send_notification(
            self.api_retry_max, self.api_retry_interval, self.event)

        mock_conn.instance_ha.create_notification.assert_called_once_with(
            type=self.event['notification']['type'],
            hostname=self.event['notification']['hostname'],
            generated_time=self.event['notification']['generated_time'],
            payload=self.event['notification']['payload'])

    @mock.patch.object(eventlet.greenthread, 'sleep')
    @mock.patch.object(connection, 'Connection')
    @mock.patch.object(ks_loading, 'load_session_from_conf_options')
    @mock.patch.object(ks_loading, 'load_auth_from_conf_options')
    def test_send_notification_500_error(
        self, mock_auth, mock_session, mock_connection, mock_sleep):

        mock_conn = mock.Mock()
        mock_conn.instance_ha.return_value = mock.Mock()
        mock_conn.instance_ha.create_notification.return_value = mock.Mock()
        mock_connection.return_value = mock_conn

        response = FakeResponse(status_code=500)
        status_ex = exceptions.HttpException(response=response)

        mock_conn.instance_ha.create_notification.side_effect = status_ex
        mock_sleep.return_value = None

        notifier = masakari.SendNotification()
        notifier.send_notification(
            self.api_retry_max, self.api_retry_interval, self.event)

        mock_conn.instance_ha.create_notification.assert_called_with(
            type=self.event['notification']['type'],
            hostname=self.event['notification']['hostname'],
            generated_time=self.event['notification']['generated_time'],
            payload=self.event['notification']['payload'])
        self.assertEqual(self.api_retry_max + 1,
                         mock_conn.instance_ha.create_notification.call_count)
